#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ts=4 sw=4 et
"""
================================
K40 Calibration Batch Processing
================================

Standalone job submitter for K40 offline calibrations with KM3Pipe.

.. code-block:: console

    Usage:
        qk40calib.py OUTPUT_PATH [options]
        qk40calib.py (-h | --help)

    Options:
        OUTPUT_PATH  Folder to store the calibration data.
        -d DET_ID    Detector ID [default: 29].
        -t TMAX      Coincidence time window [default: 10].
        -n N_RUNS    Number of runs to process per job [default: 10].
        -e ET        Estimated walltime per run in minutes [default: 8].
        -m VMEM      Estimated vmem for a job [default: 8G].
        -s RUNSETUP  Match [default: PHYS.1710v5-TUNED.HRV19.3D_T_S_MX.NBMODULE].
        -j JOBNAME   The name of the submitted jobs [default: k40calib].
        -l LOG_PATH  Path of the job log files [default: qlogs].
        -q           Dryrun: don't submit jobs, just print the first job script.
        -h --help    Show this screen.

"""

import os
import re
from glob import glob
import time

import km3db
from km3pipe.shell import qsub, Script
import km3pipe as kp
from docopt import docopt


def main():
    args = docopt(__doc__)

    DET_ID = int(args["-d"])
    TMAX = int(args["-t"])
    ET_PER_RUN = int(args["-e"]) * 60  # [s]
    RUNS_PER_JOB = int(args["-n"])
    VMEM = args["-m"]
    CWD = os.getcwd()
    LOG_PATH = args["-l"]
    JOB_NAME = args["-j"]
    CALIB_PATH = os.path.join(CWD, args["OUTPUT_PATH"])
    RUN_SUBSTR = args["-s"]
    DRYRUN = args["-q"]

    if not os.path.exists(CALIB_PATH):
        os.makedirs(CALIB_PATH)

    run_table = km3db.StreamDS(container="pd").get("run_table", det_id=DET_ID)
    phys_run_table = run_table[run_table.RUNSETUPNAME.str.contains(RUN_SUBSTR)]
    phys_runs = set(phys_run_table.RUN)
    processed_runs = set(
        int(re.search("_\\d{8}_(\\d{8})", s).group(1))
        for s in glob(os.path.join(CALIB_PATH, "*.k40_cal.p"))
    )
    remaining_runs = list(phys_runs - processed_runs)
    print("Remaining runs: {}".format(remaining_runs))
    s = Script()

    for job_id, runs_chunk in enumerate(kp.tools.chunks(remaining_runs, RUNS_PER_JOB)):
        n_runs = len(runs_chunk)
        print("Preparing batch script for a chunk of {} runs.".format(len(runs_chunk)))
        s.add("cd $TMPDIR; mkdir -p $USER; cd $USER")
        for run in runs_chunk:
            s.add("echo Processing {}:".format(run))
            irods_path = kp.tools.irods_path(DET_ID, run)
            root_filename = os.path.basename(irods_path)
            calib_filename = root_filename + ".k40_cal.p"
            s.add("iget -v {}".format(irods_path))
            s.add(
                "CTMIN=$(JPrint -f {}|grep '^ctMin'|awk '{{print $2}}')".format(
                    root_filename
                )
            )
            s.add(
                "k40calib {} {} -t {} -c $CTMIN -o {}".format(
                    root_filename, DET_ID, TMAX, calib_filename
                )
            )
            s.add("cp {} {}".format(calib_filename, CALIB_PATH))
            s.add("rm -f {}".format(root_filename))
            s.add("rm -f {}".format(calib_filename))
            s.add("echo Run {} processed.".format(run))
            s.add("echo " + 42 * "=")

        walltime = time.strftime("%H:%M:%S", time.gmtime(ET_PER_RUN * n_runs))
        qsub(
            s,
            "{}_{}".format(JOB_NAME, job_id),
            walltime=walltime,
            vmem=VMEM,
            log_path=LOG_PATH,
            irods=True,
            dryrun=DRYRUN,
        )

        if DRYRUN:
            break

        s.clear()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Filename: tres.py
"""
A tool to calculate and extract Cherenkov hit time residuals from reconstruced
events in offline ROOT files (currently only the JMuon chain).

Usage:
    tres [options] FILENAME
    tres (-h | --help)
    tres --version

Options:
    -o OUTFILE          Output file.
    --single-muon-only  Filter out single muon events
                        (only possible for RBR atmospheric muons - MUPAGE)
    -d DETX             Use a detector file to calibrate hits (not needed for
                        offline files created with km3net-dataformat
                        version 2+).
    -n N_EVENTS         Number of events to extract.
    -h --help           Show this screen.
    --version           Show the version.

"""


def run_pipeline(args):
    import km3pipe as kp
    import km3modules as km
    import km3io

    fname = args["FILENAME"]

    if args["-o"] is None:
        outfile = fname + ".tres.h5"
    else:
        outfile = args["-o"]

    def single_muon_filter(blob):
        """Only let events pass with exactly one muon"""
        if blob["event"].n_mc_tracks == 1:
            return blob

    def reco_filter(blob):
        """Filter out events without reco tracks"""
        if blob["event"].n_tracks > 0:
            return blob

    class TimeResidualsCalculator(kp.Module):
        def configure(self):
            detx = self.get("detx", default=None)
            self.calib = None

            if detx is not None:
                self.calib = kp.calib.Calibration(filename=detx)

        def process(self, blob):
            event = blob["event"]
            hits = event.hits

            if self.calib is not None:
                hits = self.calib.apply(hits)
                times = hits.time  # TODO: harmonise names
            else:
                times = hits.t  # TODO: harmonise names

            bt = km3io.tools.best_jmuon(event.tracks)
            cherenkov_params = kp.physics.cherenkov(hits, bt)

            t_res = kp.NDArray(
                times - cherenkov_params["t_photon"],
                h5loc="/t_res",
                title="Cherenkov hit time residuals",
            )
            blob["TRes"] = t_res

            return blob

    pipe = kp.Pipeline()
    pipe.attach(km.StatusBar, every=1000)
    pipe.attach(kp.io.OfflinePump, filename=fname)
    if args["--single-muon-only"]:
        pipe.attach(single_muon_filter)
    pipe.attach(reco_filter)
    pipe.attach(TimeResidualsCalculator, detx=args["-d"])
    pipe.attach(km.Keep, keys=["TRes"])
    pipe.attach(kp.io.HDF5Sink, filename=outfile)
    if args["-n"] is not None:
        pipe.drain(int(args["-n"]))
    else:
        pipe.drain()


def main():

    from docopt import docopt

    run_pipeline(docopt(__doc__))


if __name__ == "__main__":

    main()



#!/usr/bin/env python3
from collections import defaultdict

import numpy as np
import awkward as ak

import km3pipe as kp
import km3io

USR_MC_TRACKS_KEYS = [b"energy_lost_in_can", b"bx", b"by", b"ichan", b"cc"]


class HitsTabulator(kp.Module):
    """
    Create `kp.Table` from hits provided by `km3io`.

    Parameters
    ----------
    kind: str
      The kind of hits to tabulate:
        "offline": the hits in an offline file
        "mc": MC hits
    split: bool (default: True)
      Defines whether the hits should be split up into individual arrays
      in a single group (e.g. hits/dom_id, hits/channel_id) or stored
      as a single HDF5Compound array (e.g. hits).
    """

    def configure(self):
        self.kind = self.require("kind")
        self.with_calibration = self.get("with_calibration", default=False)
        self.split = self.get("split", default=True)

    def process(self, blob):
        if self.kind == "offline":
            n = blob["event"].n_hits
            if n == 0:
                return blob
            hits = blob["event"].hits

            hits_data = {
                "channel_id": hits.channel_id,
                "dom_id": hits.dom_id,
                "time": hits.t,
                "tot": hits.tot,
                "triggered": hits.trig,
            }

            if self.with_calibration:
                hits_data["pos_x"] = hits.pos_x
                hits_data["pos_y"] = hits.pos_y
                hits_data["pos_z"] = hits.pos_z
                hits_data["dir_x"] = hits.dir_x
                hits_data["dir_y"] = hits.dir_y
                hits_data["dir_z"] = hits.dir_z
                hits_data["tdc"] = hits.tdc

            blob["Hits"] = kp.Table(
                hits_data,
                h5loc="/hits",
                split_h5=self.split,
                name="Hits",
            )

        if self.kind == "mc":
            n = blob["event"].n_mc_hits
            if n == 0:
                return blob
            mc_hits = blob["event"].mc_hits
            blob["McHits"] = kp.Table(
                {
                    "a": mc_hits.a,
                    "origin": mc_hits.origin,
                    "pmt_id": mc_hits.pmt_id,
                    "time": mc_hits.t,
                },
                h5loc="/mc_hits",
                split_h5=self.split,
                name="McHits",
            )

        return blob


class MCTracksTabulator(kp.Module):
    """
    Create `kp.Table` from MC tracks provided by `km3io`.

    Parameters
    ----------
    split: bool (default: False)
      Defines whether the tracks should be split up into individual arrays
      in a single group (e.g. mc_tracks/by, mc_tracks/origin) or stored
      as a single HDF5Compound array (e.g. mc_tracks).
    read_usr_data: bool (default: False)
      Parses usr-data which is originally meant for user stored values, but
      was abused by generator software to store properties. This issue will
      be sorted out hopefully soon as it dramatically decreases the processing
      performance and usability.
    """

    def configure(self):
        self.split = self.get("split", default=False)

        self._read_usr_data = self.get("read_usr_data", default=False)
        if self._read_usr_data:
            self.log.warning(
                "Reading usr-data will massively decrease the performance."
            )

    def process(self, blob):
        n = blob["event"].n_mc_tracks
        if n == 0:
            return blob

        mc_tracks = blob["event"].mc_tracks
        blob["McTracks"] = self._parse_mc_tracks(mc_tracks)
        return blob

    def _parse_usr_to_dct(self, mc_tracks):
        dct = defaultdict(list)
        for k in USR_MC_TRACKS_KEYS:
            dec_key = k.decode("utf_8")
            for i in range(len(mc_tracks.usr_names)):
                value = np.nan
                if k in mc_tracks.usr_names[i]:
                    mask = mc_tracks.usr_names[i] == k
                    value = mc_tracks.usr[i][mask][0]
                dct[dec_key].append(value)
        return dct

    def _parse_mc_tracks(self, mc_tracks):
        dct = {
            "dir_x": mc_tracks.dir_x,
            "dir_y": mc_tracks.dir_y,
            "dir_z": mc_tracks.dir_z,
            "pos_x": mc_tracks.pos_x,
            "pos_y": mc_tracks.pos_y,
            "pos_z": mc_tracks.pos_z,
            "energy": mc_tracks.E,
            "time": mc_tracks.t,
            "pdgid": mc_tracks.pdgid,
            "id": mc_tracks.id,
            "length": mc_tracks.len,
        }
        if self._read_usr_data:
            dct.update(self._parse_usr_to_dct(mc_tracks))
        return kp.Table(dct, name="McTracks", h5loc="/mc_tracks", split_h5=self.split)


class RecoTracksTabulator(kp.Module):
    """
    Create `kp.Table` from recostruced tracks provided by `km3io`.

    Parameters
    ----------
    best_tracks: bool (default: False)
      Additionally determine best track.
    split: bool (default: False)
      Defines whether the tracks should be split up into individual arrays
      in a single group (e.g. reco/tracks/dom_id, reco/tracks/channel_id) or stored
      as a single HDF5Compound array (e.g. reco/tracks).
    """

    def configure(self):

        self.split = self.get("split", default=False)
        self.best_tracks = self.get("best_tracks", default=False)
        self.aashower_legacy = self.get("aashower_legacy", default=False)

        self._best_track_fmap = {
            km3io.definitions.reconstruction.JMUONPREFIT: (
                km3io.tools.best_jmuon,
                "best_jmuon",
            ),
            km3io.definitions.reconstruction.JSHOWERPREFIT: (
                km3io.tools.best_jshower,
                "best_jshower",
            ),
            km3io.definitions.reconstruction.DUSJSHOWERPREFIT: (
                km3io.tools.best_dusjshower,
                "best_dusjshower",
            ),
            km3io.definitions.reconstruction.AASHOWERFITPREFIT: (
                km3io.tools.best_aashower,
                "best_aashower",
            ),
        }

    def process(self, blob):
        n_tracks = blob["event"].n_tracks
        # we first check if there are any tracks, otherwise the other calls will raise
        if n_tracks == 0:
            return blob

        all_tracks = blob["event"].tracks

        if self.aashower_legacy == True:

            all_tracks.rec_stages = np.where(
                all_tracks.rec_type
                == km3io.definitions.reconstruction.AANET_RECONSTRUCTION_TYPE,
                all_tracks.rec_stages + 300,
                all_tracks.rec_stages,
            )

        # put all tracks into the blob
        self._put_tracks_into_blob(blob, all_tracks, "tracks", n_tracks)

        # select the best track using the km3io tools
        if self.best_tracks:

            # check if it contains any of the specific reco types (can be several)
            for stage, (best_track, reco_name) in self._best_track_fmap.items():
                if ak.any(stage == all_tracks.rec_stages):
                    tracks = best_track(all_tracks)
                    self._put_tracks_into_blob(blob, tracks, reco_name, 1)

        return blob

    def _put_tracks_into_blob(self, blob, tracks, reco_identifier, n_tracks):

        """
        Put a certain type of "tracks" in the blob and give specific name.

        Parameters
        ----------
        tracks : awkward array
            The tracks object to be put in the blob eventually. Can be only best tracks.
        identifier : string
            A string to name the kp table.
        n_tracks : int
            The number of tracks from before. Use to distinguish between best and all tracks.

        """

        reco_tracks = dict(
            pos_x=tracks.pos_x,
            pos_y=tracks.pos_y,
            pos_z=tracks.pos_z,
            dir_x=tracks.dir_x,
            dir_y=tracks.dir_y,
            dir_z=tracks.dir_z,
            E=tracks.E,
            rec_type=tracks.rec_type,
            t=tracks.t,
            likelihood=tracks.lik,
            length=tracks.len,  # do all recos have this?
        )

        if n_tracks != 1:
            reco_tracks.update(
                id=tracks.id,
                idx=np.arange(n_tracks),
            )

        n_columns = max(km3io.definitions.fitparameters.values()) + 1
        fitinf_array = np.ma.filled(
            ak.to_numpy(ak.pad_none(tracks.fitinf, target=n_columns, axis=-1)),
            fill_value=np.nan,
        ).astype("float32")
        fitinf_split = np.split(fitinf_array, fitinf_array.shape[-1], axis=-1)

        if n_tracks == 1:
            for fitparam, idx in km3io.definitions.fitparameters.items():
                reco_tracks[fitparam] = fitinf_split[idx][0]

        else:
            for fitparam, idx in km3io.definitions.fitparameters.items():
                reco_tracks[fitparam] = fitinf_split[idx][:, 0]

        blob["Reco_" + reco_identifier] = kp.Table(
            reco_tracks,
            h5loc=f"/reco/" + reco_identifier,
            name="Reco " + reco_identifier,
            split_h5=self.split,
        )

        # write out the rec stages only once with all tracks
        if n_tracks != 1:

            _rec_stage = np.array(
                ak.highlevel.Array(ak.flatten(tracks.rec_stages)._layout)
            )
            _counts = ak.count(tracks.rec_stages, axis=1)
            _idx = np.repeat(np.arange(n_tracks), _counts)

            blob["RecStages"] = kp.Table(
                dict(rec_stage=_rec_stage, idx=_idx),
                # Just to save space, we specify smaller dtypes.
                # We assume there will be never more than 32767
                # reco tracks for a single reconstruction type.
                dtypes=[("rec_stage", np.int16), ("idx", np.uint16)],
                h5loc=f"/reco/rec_stages",
                name="Reconstruction Stages",
                split_h5=self.split,
            )


class EventInfoTabulator(kp.Module):
    """
    Create `kp.Table` from event information provided by `km3io`.

    """

    def process(self, blob):

        # get the sim program
        if blob["header"]:
            if "simul" in blob["header"].keys():
                sim_program = blob["header"].simul.program
            else:  # not existent for real data
                sim_program = None
        else:
            sim_program = None

        blob["EventInfo"] = self._parse_eventinfo(blob["event"], sim_program)
        return blob

    def _parse_eventinfo(self, event, sim_program):
        wgt1, wgt2, wgt3, wgt4 = self._parse_wgts(event.w)
        tab_data = {
            "event_id": event.id,
            "run_id": event.run_id,
            "weight_w1": wgt1,
            "weight_w2": wgt2,
            "weight_w3": wgt3,
            "weight_w4": wgt4,
            "timestamp": event.t_sec,
            "nanoseconds": event.t_ns,
            "mc_time": event.mc_t,
            "trigger_mask": event.trigger_mask,
            "trigger_counter": event.trigger_counter,
            "overlays": event.overlays,
            "det_id": event.det_id,
            "frame_index": event.frame_index,
            "mc_run_id": event.mc_run_id,
        }

        if sim_program != None:

            # unfold the info in the w2list
            w2list_dict = self._unfold_w2list(event.w2list, sim_program)
            tab_data.update(w2list_dict)

        info = kp.Table(tab_data, h5loc="/event_info", name="EventInfo")
        return info

    def _unfold_w2list(self, w2list, sim_program):

        w2list_dict = {}
        definitions_dict = {}

        if sim_program.lower() == "gseagen":
            definitions_dict = km3io.definitions.w2list_gseagen
        elif sim_program.lower() == "genhen":
            definitions_dict = km3io.definitions.w2list_genhen
        # for cases like sim_program == "MUPAGE", the w2list is empty

        for key, idx in definitions_dict.items():
            w2list_dict[key] = np.nan if idx >= len(w2list) else w2list[idx]

        return w2list_dict

    @staticmethod
    def _parse_wgts(wgt):
        if len(wgt) == 3:
            wgt1, wgt2, wgt3 = wgt
            wgt4 = np.nan
        elif len(wgt) == 4:
            # what the hell is w4?
            wgt1, wgt2, wgt3, wgt4 = wgt
        else:
            wgt1 = wgt2 = wgt3 = wgt4 = np.nan
        return wgt1, wgt2, wgt3, wgt4


class OfflineHeaderTabulator(kp.Module):
    def process(self, blob):
        if blob["header"]:
            blob["RawHeader"] = kp.io.hdf5.header2table(blob["header"])
        return blob



import urwid
import random


class ItemWidget(urwid.WidgetWrap):
    def __init__(self, id, description):
        self.id = id
        self.content = "item %s: %s..." % (str(id), description[:25])
        self.item = [
            (
                "fixed",
                15,
                urwid.Padding(
                    urwid.AttrWrap(urwid.Text("item %s" % str(id)), "body", "focus"),
                    left=2,
                ),
            ),
            urwid.AttrWrap(urwid.Text("%s" % description), "body", "focus"),
        ]
        w = urwid.Columns(self.item)
        self.__super.__init__(w)

    def selectable(self):
        return True

    def keypress(self, size, key):
        return key


def main():

    palette = [
        ("body", "dark cyan", "", "standout"),
        ("focus", "dark red", "", "standout"),
        ("head", "light red", "black"),
    ]

    lorem = [
        "Lorem ipsum dolor sit amet, consectetur adipiscing elit.",
        "Sed sollicitudin, nulla id viverra pulvinar.",
        "Cras a magna sit amet felis fringilla lobortis.",
    ]

    def keystroke(input):
        if input in ("q", "Q"):
            raise urwid.ExitMainLoop()

        if input is "enter":
            focus = listbox.get_focus()[0].content
            view.set_header(
                urwid.AttrWrap(urwid.Text("selected: %s" % str(focus)), "head")
            )

    items = []
    for i in range(100):
        items.append(ItemWidget(i, random.choice(lorem)))

    header = urwid.AttrMap(urwid.Text("selected:"), "head")
    listbox = urwid.ListBox(urwid.SimpleListWalker(items))
    view = urwid.Frame(urwid.AttrWrap(listbox, "body"), header=header)
    loop = urwid.MainLoop(view, palette, unhandled_input=keystroke)
    loop.run()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""

==================
PMT Directions
==================

Plot the PMT directions for a given DOM.

"""

# Author: Tamas Gal <tgal@km3net.de>
# License: BSD-3
from mpl_toolkits.mplot3d import axes3d
import matplotlib.pyplot as plt

from km3net_testdata import data_path

import km3pipe as kp
import km3pipe.style

km3pipe.style.use("km3pipe")
km3pipe.style.use("talk")

detx = data_path(
    "detx/orca_115strings_av23min20mhorizontal_18OMs_alt9mvertical_v1.detx"
)
det = kp.hardware.Detector(detx)

#####################################################
# Alternatively, you can use the `det_id` to retrieve the geometry from the DB.

# det = kp.hardware.Detector(det_id=29)

#####################################################
# Let's take the first DOM ID
dom_id = det.dom_ids[0]

#####################################################
# ...and get the table of the PMTs in the chosen DOM:

pmts = det.pmts[det.pmts.dom_id == dom_id]

#####################################################
# The `quiver` function can directly plot the PMT data, since those are
# stored as numpy arrays.
fig = plt.figure()
ax = plt.axes(projection="3d")

ax.quiver(
    pmts.pos_x, pmts.pos_y, pmts.pos_z, pmts.dir_x, pmts.dir_y, pmts.dir_z, length=0.1
)

ax.set_xlabel("x [m]")
ax.set_ylabel("y [m]")
ax.set_zlabel("z [m]")

ax.set_title("PMT directions on DOM {}".format(dom_id))



# -*- coding: utf-8 -*-
"""
=============================
Reading and Parsing EVT files
=============================

This example shows how to read and parse EVT files, which are used in our
Monte Carlo productions.
"""

# Author: Tamas Gal <tgal@km3net.de>, Moritz Lotze >mlotze@km3net.de>
# License: BSD-3
import matplotlib.pyplot as plt
import numpy as np

import km3pipe as kp
import km3modules as km
from km3net_testdata import data_path

kp.style.use("km3pipe")

filename = data_path("evt/example_numuCC.evt")
detx = data_path("detx/km3net_jul13_90m_r1494_corrected.detx")


class VertexHitDistanceCalculator(kp.Module):
    """Calculate vertex-hit-distances"""

    def configure(self):
        self.distances = []

    def process(self, blob):
        tracks = blob["TrackIns"]
        muons = tracks[tracks.type == 5]
        muon = kp.Table(muons[np.argmax(muons.energy)])
        hits = blob["CalibHits"]
        dist = kp.math.pld3(hits.pos, muon.pos, muon.dir)
        self.distances.append(dist)
        return blob

    def finish(self):
        dist_flat = np.concatenate(self.distances)
        plt.hist(dist_flat)
        plt.xlabel("distance between hits and muon / m")
        plt.ylabel("count")
        plt.tight_layout()
        plt.savefig("dists.png")


pipe = kp.Pipeline()
pipe.attach(kp.io.EvtPump, filename=filename, parsers=["km3"])
pipe.attach(km.StatusBar, every=100)
pipe.attach(kp.calib.Calibration, filename=detx)
pipe.attach(VertexHitDistanceCalculator)
pipe.drain(5)



# Filename: common.py
# -*- coding: utf-8 -*-
# pylint: disable=locally-disabled
"""
A collection of commonly used modules.

"""

import sqlite3
from time import time

import numpy as np

import km3pipe as kp
from km3pipe import Module, Blob
from km3pipe.tools import prettyln
from km3pipe.sys import peak_memory_usage

log = kp.logger.get_logger(__name__)


class Dump(Module):
    """Print the content of the blob.

    Parameters
    ----------
    keys: collection(string), optional [default=None]
        Keys to print. If None, print all keys.
    full: bool, default=False
        Print blob values too, not just the keys?
    """

    def configure(self):
        self.keys = self.get("keys") or None
        self.full = self.get("full") or False
        key = self.get("key") or None
        if key and not self.keys:
            self.keys = [key]

    def process(self, blob):
        keys = sorted(blob.keys()) if self.keys is None else self.keys
        for key in keys:
            print(key + ":")
            if self.full:
                print(blob[key].__repr__())
            print("")
        print("----------------------------------------\n")
        return blob


class Delete(Module):
    """Remove specific keys from the blob.

    Parameters
    ----------
    keys: collection(string), optional
        Keys to remove.
    """

    def configure(self):
        self.keys = self.get("keys") or set()
        key = self.get("key") or None
        if key and not self.keys:
            self.keys = [key]

    def process(self, blob):
        for key in self.keys:
            blob.pop(key, None)
        return blob


class Keep(Module):
    """Keep only specified keys in the blob.

    Parameters
    ----------
    keys: collection(string), optional
        Keys to keep. Everything else is removed.
    """

    def configure(self):
        self.keys = self.get("keys", default=set())
        key = self.get("key", default=None)
        self.h5locs = self.get("h5locs", default=set())
        if key and not self.keys:
            self.keys = [key]

    def process(self, blob):
        out = Blob()
        for key in blob.keys():
            if key in self.keys:
                out[key] = blob[key]
            elif hasattr(blob[key], "h5loc") and blob[key].h5loc.startswith(
                tuple(self.h5locs)
            ):
                out[key] = blob[key]
        return out


class HitCounter(Module):
    """Prints the number of hits"""

    def process(self, blob):
        try:
            self.cprint("Number of hits: {0}".format(len(blob["Hit"])))
        except KeyError:
            pass
        return blob


class HitCalibrator(Module):
    """A very basic hit calibrator, which requires a `Calibration` module."""

    def configure(self):
        self.input_key = self.get("input_key", default="Hits")
        self.output_key = self.get("output_key", default="CalibHits")

    def process(self, blob):
        if self.input_key not in blob:
            self.log.warn("No hits found in key '{}'.".format(self.input_key))
            return blob
        hits = blob[self.input_key]
        chits = self.calibration.apply(hits)
        blob[self.output_key] = chits
        return blob


class BlobIndexer(Module):
    """Puts an incremented index in each blob for the key 'blob_index'"""

    def configure(self):
        self.blob_index = 0

    def process(self, blob):
        blob["blob_index"] = self.blob_index
        self.blob_index += 1
        return blob


class StatusBar(Module):
    """Displays the current blob number."""

    def configure(self):
        self.iteration = 1

    def process(self, blob):
        prettyln("Blob {0:>7}".format(self.every * self.iteration))
        self.iteration += 1
        return blob

    def finish(self):
        prettyln(".", fill="=")


class TickTock(Module):
    """Display the elapsed time.

    Parameters
    ----------
    every: int, optional [default=1]
        Number of iterations between printout.
    """

    def configure(self):
        self.t0 = time()

    def process(self, blob):
        t1 = (time() - self.t0) / 60
        prettyln("Time/min: {0:.3f}".format(t1))
        return blob


class MemoryObserver(Module):
    """Shows the maximum memory usage

    Parameters
    ----------
    every: int, optional [default=1]
        Number of iterations between printout.
    """

    def process(self, blob):
        memory = peak_memory_usage()
        prettyln("Memory peak: {0:.3f} MB".format(memory))
        return blob


class Siphon(Module):
    """A siphon to accumulate a given volume of blobs.

    Parameters
    ----------
    volume: int
      number of blobs to hold
    flush: bool
      discard blobs after accumulation

    """

    def configure(self):
        self.volume = self.require("volume")  # [blobs]
        self.flush = self.get("flush", default=False)

        self.blob_count = 0

    def process(self, blob):
        self.blob_count += 1
        if self.blob_count > self.volume:
            log.debug("Siphone overflow reached!")
            if self.flush:
                log.debug("Flushing the siphon.")
                self.blob_count = 0
            return blob


class MultiFilePump(kp.Module):
    """Use the given pump to iterate through a list of files.

    The group_id will be reset so that it's unique for each iteration.

    Parameters
    ----------
    pump: Pump
      The pump to be used to generate the blobs.
    filenames: iterable(str)
      List of filenames.
    kwargs: dict(str -> any) optional
      Keyword arguments to be passed to the pump.

    """

    def configure(self):
        self.pump = self.require("pump")
        self.filenames = self.require("filenames")
        self.kwargs = self.get("kwargs", default={})
        self.blobs = self.blob_generator()
        self.cprint("Iterating through {} files.".format(len(self.filenames)))
        self.n_processed = 0
        self.group_id = 0

    def blob_generator(self):
        for filename in self.filenames:
            self.cprint("Current file: {}".format(filename))
            pump = self.pump(filename=filename, **self.kwargs)
            for blob in pump:
                self._set_group_id(blob)
                blob["filename"] = filename
                yield blob
                self.group_id += 1
            self.n_processed += 1

    def _set_group_id(self, blob):
        for key, entry in blob.items():
            if isinstance(entry, kp.Table):
                if hasattr(entry, "group_id"):
                    entry.group_id = self.group_id
                else:
                    blob[key] = entry.append_columns("group_id", self.group_id)

    def process(self, blob):
        return next(self.blobs)

    def finish(self):
        self.cprint(
            "Fully processed {} out of {} files.".format(
                self.n_processed, len(self.filenames)
            )
        )


class LocalDBService(kp.Module):
    """Provides a local sqlite3 based database service to store information"""

    def configure(self):
        self.filename = self.require("filename")
        self.thread_safety = self.get("thread_safety", default=True)
        self.connection = None

        self.expose(self.create_table, "create_table")
        self.expose(self.table_exists, "table_exists")
        self.expose(self.insert_row, "insert_row")
        self.expose(self.query, "query")

        self._create_connection()

    def _create_connection(self):
        """Create database connection"""
        try:
            self.connection = sqlite3.connect(
                self.filename, check_same_thread=self.thread_safety
            )
            self.cprint(sqlite3.version)
        except sqlite3.Error as exception:
            self.log.error(exception)

    def query(self, query):
        """Execute a SQL query and return the result of fetchall()"""
        cursor = self.connection.cursor()
        cursor.execute(query)
        return cursor.fetchall()

    def insert_row(self, table, column_names, values):
        """Insert a row into the table with a given list of values"""
        cursor = self.connection.cursor()
        query = "INSERT INTO {} ({}) VALUES ({})".format(
            table, ", ".join(column_names), ",".join("'" + str(v) + "'" for v in values)
        )
        cursor.execute(query)
        self.connection.commit()

    def create_table(self, name, columns, types, overwrite=False):
        """Create a table with given columns and types, overwrite if specified


        The `types` should be a list of SQL types, like ["INT", "TEXT", "INT"]
        """
        cursor = self.connection.cursor()
        if overwrite:
            cursor.execute("DROP TABLE IF EXISTS {}".format(name))

        cursor.execute(
            "CREATE TABLE {} ({})".format(
                name, ", ".join(["{} {}".format(*c) for c in zip(columns, types)])
            )
        )
        self.connection.commit()

    def table_exists(self, name):
        """Check if a table exists in the database"""
        cursor = self.connection.cursor()
        cursor.execute(
            "SELECT count(name) FROM sqlite_master "
            "WHERE type='table' AND name='{}'".format(name)
        )
        return cursor.fetchone()[0] == 1

    def finish(self):
        if self.connection:
            self.connection.close()


class Observer(kp.Module):
    """A simple helper to observe the blobs in a test pipeline.

    Parameters
    ----------
    count: int
      The exact number of iterations the pipeline has to drain
    required_keys: list(str)
      A list of keys which has to be present in a blob in every cycle.
    """

    def configure(self):
        self.count = self.get("count")
        self.required_keys = self.get("required_keys", default=[])
        self._count = 0

    def process(self, blob):
        self._count += 1
        for key in self.required_keys:
            assert key in blob
        return blob

    def finish(self):
        print(f"Target count={self._count}, actual count={self.count}")
        if self.count is not None:
            assert self.count == self._count


class FilePump(kp.Module):
    """A basic iterator for a list of files.

    Parameters
    ----------
    filenames: iterable(str)
      The filenames to be iterated over which are put into ``blob["filename"]``

    """

    def configure(self):
        self.filenames = self.require("filenames")
        self.blobs = self.blob_generator()

    def blob_generator(self):
        for filename in self.filenames:
            yield kp.Blob({"filename": filename})

    def process(self, blob):
        blob.update(next(self.blobs))
        return blob



# Filename: test_ch.py
# pylint: disable=locally-disabled,C0111,R0904,C0301,C0103,W0212
from km3pipe.testing import TestCase, patch, Mock
from km3pipe.io.ch import CHPump

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2018, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class TestCHPump(TestCase):
    @patch("km3pipe.io.ch.CHPump._init_controlhost")
    @patch("km3pipe.io.ch.CHPump._start_thread")
    def test_init(self, init_controlhost_mock, start_thread_mock):
        CHPump()
        init_controlhost_mock.assert_called_once()
        start_thread_mock.assert_called_once()



# Filename: calib.py
# pylint: disable=locally-disabled
"""
Calibration.

"""
import awkward as ak
import numba as nb
import numpy as np

import km3db
import km3io

from thepipe import Module
from km3pipe.hardware import Detector
from km3pipe.dataclasses import Table
from km3pipe.tools import istype
from km3pipe.logger import get_logger

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = ["Thomas Heid"]
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

log = get_logger(__name__)


class Calibration(Module):
    """A module which applies time, position and rotation corrections to hits.

    This module also calibrates MC hits, but be aware, t0s are not appended to
    the MC hit times.
    Additionally, the global PMT ID is added to regular hits as ``pmt_id`` and
    in case of MC hits, the ``dom_id`` and ``channel_id`` (DAQ) are set.

    Parameters
    ----------
    apply: bool, optional [default=True]
        Apply the calibration to the hits (add position/direction/t0)?
    filename: str, optional [default=None]
        DetX file with detector description.
    det_id: int, optional
        .detx ID of detector (when retrieving from database).
    t0set: optional
        t0set (when retrieving from database).
    calibset: optional
        calibset (when retrieving from database).
    key: str, optional [default="Hits"]
        the blob key of the hits
    outkey: str, optional [default="CalibHits"]
        the output blob key of the calibrated hits
    key_mc: str, optional [default="McHits"]
        the blob key of the MC hits (if present)
    outkey_mc: str, optional [default="CalibMcHits"]
        the output blob key of the calibrated MC hits
    """

    __name__ = "Calibration"
    name = "Calibration"

    def configure(self):
        self._should_apply = self.get("apply", default=True)
        self.filename = self.get("filename")
        self.det_id = self.get("det_id")
        self.run = self.get("run")
        self.t0set = self.get("t0set")
        self.calibset = self.get("calibset")
        self.detector = self.get("detector")
        self.key = self.get("key", default="Hits")
        self.outkey = self.get("outkey", default="CalibHits")
        self.key_mc = self.get("key_mc", default="McHits")
        self.outkey_mc = self.get("outkey_mc", default="CalibMcHits")
        self._pos_dom_channel = None
        self._dir_dom_channel = None
        self._t0_dom_channel = None
        self._pos_pmt_id = None
        self._dir_pmt_id = None
        self._t0_pmt_id = None
        self._lookup_tables = None  # for Numba

        if self.det_id and self.run:
            self.cprint(
                "Grabbing the calibration for Det ID {} and run {}".format(
                    self.det_id, self.run
                )
            )
            raw_detx = km3db.tools.detx_for_run(self.det_id, self.run)
            self.detector = Detector(string=raw_detx)
            self._create_dom_channel_lookup()
            self._create_pmt_id_lookup()
            return

        if self.filename or self.det_id:
            if self.filename is not None:
                self.detector = Detector(filename=self.filename)
            if self.det_id:
                self.detector = Detector(
                    det_id=self.det_id, t0set=self.t0set, calibset=self.calibset
                )

        if self.detector is not None:
            self.log.debug("Creating lookup tables")
            self._create_dom_channel_lookup()
            self._create_pmt_id_lookup()
        else:
            self.log.critical("No detector information loaded.")

    def process(self, blob):
        if self._should_apply:
            blob[self.outkey] = self.apply(blob[self.key])
            if self.key_mc in blob:
                blob[self.outkey_mc] = self.apply(blob[self.key_mc])
        return blob

    def get_detector(self):
        """Return the detector"""
        return self.detector

    def dus(self, hits):
        """Return the DUs for given hits"""
        dom_ids = hits["dom_id"]
        return _dus(dom_ids, len(dom_ids), self._calib_by_dom_and_channel)

    def floors(self, hits):
        """Return the floors for given hits"""
        dom_ids = hits["dom_id"]
        return _floors(dom_ids, len(dom_ids), self._calib_by_dom_and_channel)

    def apply_t0(self, hits):
        """Apply only t0s"""
        apply_t0_nb(hits.time, hits.dom_id, hits.channel_id, self._lookup_tables)
        return hits

    def apply(self, hits, no_copy=False, correct_slewing=True, slewing_variant=3):
        """Add x, y, z, t0 (and du, floor if DataFrame) columns to the hits."""
        if not no_copy:
            try:
                hits = hits.copy()
            except AttributeError:  # probably a km3io object
                pass

        if isinstance(hits, (ak.Array, ak.Record, km3io.rootio.Branch)):
            if hasattr(hits, "dom_id"):
                hits = Table(
                    dict(
                        dom_id=hits.dom_id,
                        channel_id=hits.channel_id,
                        time=hits.t,
                        tot=hits.tot,
                        triggered=hits.trig,
                    )
                )
            else:  # mc_hits in km3io
                hits = Table(
                    dict(
                        pmt_id=hits.pmt_id,
                        time=hits.t,
                        a=hits.a,
                        # TODO: Not all MC files have these two fields
                        # pure_a=hits.pure_a,
                        # pure_t=hits.pure_t,
                        origin=hits.origin,
                    )
                )

        if istype(hits, "DataFrame"):
            # do we ever see McHits here?
            hits = Table.from_template(hits, "Hits")

        is_mc = None
        if hasattr(hits, "dom_id") and hasattr(hits, "channel_id"):
            try:
                (
                    dir_x,
                    dir_y,
                    dir_z,
                    du,
                    floor,
                    pos_x,
                    pos_y,
                    pos_z,
                    t0,
                    pmt_id,
                ) = _get_calibration_for_hits(hits, self._calib_by_dom_and_channel)
            except KeyError as e:
                self.log.critical("Wrong calibration (DETX) data provided.")
                raise
            is_mc = False
        elif hasattr(hits, "pmt_id"):
            try:
                (
                    dir_x,
                    dir_y,
                    dir_z,
                    du,
                    floor,
                    pos_x,
                    pos_y,
                    pos_z,
                    t0,
                    dom_id,
                    channel_id,
                ) = _get_calibration_for_mchits(hits, self._calib_by_pmt_id)
            except KeyError as e:
                self.log.critical("Wrong calibration (DETX) data provided.")
                raise
            is_mc = True
        else:
            raise TypeError(
                "Don't know how to apply calibration to '{0}'. "
                "We need at least 'dom_id' and 'channel_id', or "
                "'pmt_id'.".format(hits.name)
            )

        if hasattr(hits, "time") and not is_mc:
            if hits.time.dtype != t0.dtype:
                time = hits.time.astype("f4") + t0.astype("f4")
                hits = hits.drop_columns(["time"])
                hits = hits.append_columns(["time"], [time])
            else:
                hits.time += t0

        hits_data = {}
        for colname in hits.dtype.names:
            hits_data[colname] = hits[colname]
        calib = {
            "dir_x": dir_x,
            "dir_y": dir_y,
            "dir_z": dir_z,
            "du": du.astype(np.uint8),
            "floor": floor.astype(np.uint8),
            "pos_x": pos_x,
            "pos_y": pos_y,
            "pos_z": pos_z,
            "t0": t0,
        }

        if is_mc:
            calib["dom_id"] = dom_id.astype(np.int32)
            calib["channel_id"] = channel_id.astype(np.int32)
        else:
            calib["pmt_id"] = pmt_id.astype(np.int32)

        hits_data.update(calib)

        if correct_slewing and not is_mc:
            hits_data["time"] -= slew(hits_data["tot"], variant=slewing_variant)
        return Table(
            hits_data, h5loc=hits.h5loc, split_h5=hits.split_h5, name=hits.name
        )

    def _create_dom_channel_lookup(self):
        data = nb.typed.Dict.empty(
            key_type=nb.types.i8, value_type=nb.types.float64[:, :]
        )
        for pmt in self.detector.pmts:
            if pmt.dom_id not in data:
                data[pmt.dom_id] = np.zeros((31, 10))
            data[pmt.dom_id][pmt.channel_id] = np.asarray(
                [
                    pmt.pos_x,
                    pmt.pos_y,
                    pmt.pos_z,
                    pmt.dir_x,
                    pmt.dir_y,
                    pmt.dir_z,
                    pmt.t0,
                    pmt.du,
                    pmt.floor,
                    pmt.pmt_id,
                ],
                dtype=np.float64,
            )
        self._calib_by_dom_and_channel = data
        self._lookup_tables = [(dom, cal) for dom, cal in data.items()]

    def _create_pmt_id_lookup(self):
        data = nb.typed.Dict.empty(key_type=nb.types.i8, value_type=nb.types.float64[:])
        for pmt in self.detector.pmts:
            data[pmt.pmt_id] = np.asarray(
                [
                    pmt.pos_x,
                    pmt.pos_y,
                    pmt.pos_z,
                    pmt.dir_x,
                    pmt.dir_y,
                    pmt.dir_z,
                    pmt.t0,
                    pmt.du,
                    pmt.floor,
                    pmt.dom_id,
                    pmt.channel_id,
                ],
                dtype=np.float64,
            )
        self._calib_by_pmt_id = data

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return "Calibration: det_id({0})".format(self.det_id)


@nb.njit
def apply_t0_nb(times, dom_ids, channel_ids, lookup_tables):
    """Apply t0s using a lookup table of tuples (dom_id, calib)"""
    dom_id = 0
    lookup = np.empty((31, 9))
    for i in range(len(times)):
        cur_dom_id = dom_ids[i]
        if cur_dom_id != dom_id:
            dom_id = cur_dom_id
            for (d, m) in lookup_tables:
                if d == dom_id:
                    np.copyto(lookup, m)
        t0 = lookup[channel_ids[i]][6]
        times[i] += t0


@nb.jit(nopython=True)
def _get_calibration_for_hits(hits, lookup):
    """Append the position, direction and t0 columns and add t0 to time"""
    n = len(hits)
    cal = np.empty((n, 10))
    for i in range(n):
        calib = lookup[hits["dom_id"][i]][hits["channel_id"][i]]
        cal[i] = calib
    dir_x = cal[:, 3]
    dir_y = cal[:, 4]
    dir_z = cal[:, 5]
    du = cal[:, 7]
    floor = cal[:, 8]
    pos_x = cal[:, 0]
    pos_y = cal[:, 1]
    pos_z = cal[:, 2]
    pmt_id = cal[:, 9]

    t0 = cal[:, 6]

    return [dir_x, dir_y, dir_z, du, floor, pos_x, pos_y, pos_z, t0, pmt_id]


@nb.jit(nopython=True)
def _get_calibration_for_mchits(hits, lookup):
    """Append the position, direction and t0 columns and add t0 to time"""
    n_hits = len(hits)
    cal = np.empty((n_hits, 11))
    for i in range(n_hits):
        cal[i] = lookup[hits["pmt_id"][i]]
    dir_x = cal[:, 3]
    dir_y = cal[:, 4]
    dir_z = cal[:, 5]
    du = cal[:, 7]
    floor = cal[:, 8]
    pos_x = cal[:, 0]
    pos_y = cal[:, 1]
    pos_z = cal[:, 2]
    t0 = cal[:, 6]
    dom_id = cal[:, 9]
    channel_id = cal[:, 10]

    return [dir_x, dir_y, dir_z, du, floor, pos_x, pos_y, pos_z, t0, dom_id, channel_id]


@nb.njit
def _dus(dom_ids, n, lookup):
    dus = np.empty(n, dtype=np.uint8)
    for i in range(n):
        # DU is at index 7
        dus[i] = lookup[dom_ids[i]][0][7]  # looking only at channel ID 0
    return dus


@nb.njit
def _floors(dom_ids, n, lookup):
    dus = np.empty(n, dtype=np.uint8)
    for i in range(n):
        # floor is at index 8
        dus[i] = lookup[dom_ids[i]][0][8]  # looking only at channel ID 0
    return dus


class CalibrationService(Module):
    """A service which provides calibration routines for hits

    Parameters
    ----------
    filename: str, optional [default=None]
        DetX file with detector description.
    det_id: int, optional
        .detx ID of detector (when retrieving from database).
    t0set: optional
        t0set (when retrieving from database).
    calibset: optional
        calibset (when retrieving from database).
    detector: kp.hardware.Detector, optional
    """

    __name__ = "Calibration"
    name = "Calibration"

    def configure(self):
        self.filename = self.get("filename")
        self.det_id = self.get("det_id")
        self.t0set = self.get("t0set")
        self.calibset = self.get("calibset")

        self._detector = self.get("detector")

        if self._detector is not None:
            self._calibration = Calibration(detector=self._detector)

        self._calibration = None

        self.expose(self.calibrate, "calibrate")
        self.expose(self.get_detector, "get_detector")
        self.expose(self.get_calibration, "get_calibration")
        self.expose(self.load_calibration, "load_calibration")
        self.expose(self.correct_slewing, "correct_slewing")

    def load_calibration(self, filename=None, det_id=None, t0set=None, calibset=None):
        """Load another calibration"""
        self.filename = filename
        self.det_id = det_id
        self.t0set = t0set
        self.calibset = calibset
        self._detector = None
        self._calibration = None

    def calibrate(self, hits, correct_slewing=True):
        return self.calibration.apply(hits, correct_slewing=correct_slewing)

    @property
    def detector(self):
        if self._detector is None:
            self._detector = self.calibration.detector
        return self._detector

    def get_detector(self):
        """Extra getter to be as lazy as possible (expose triggers otherwise"""
        return self.detector

    @property
    def calibration(self):
        if self._calibration is None:
            self._calibration = Calibration(
                filename=self.filename,
                det_id=self.det_id,
                t0set=self.t0set,
                calibset=self.calibset,
            )
        return self._calibration

    def get_calibration(self):
        """Extra getter to be as lazy as possible (expose triggers otherwise"""
        return self.calibration

    def correct_slewing(self, hits):
        """Apply time slewing correction to the hit times"""
        hits.time -= slew(hits.tot)


def slew(tot, variant=3):
    """Calculate the time slewing of a PMT response for a given ToT


    Parameters
    ----------
    tot: int or np.array(int)
        Time over threshold value of a hit
    variant: int, optional
        The variant of the slew calculation.
        1: The first parametrisation approach
        2: Jannik's improvement of the parametrisation
        3: The latest lookup table approach based on lab measurements.

    Returns
    -------
    time: int
        Time slewing, which has to be subtracted from the original hit time.
    """
    if variant == 1:
        return _slew_parametrised(7.70824, 0.00879447, -0.0621101, -1.90226, tot)
    if variant == 2:
        return _slew_parametrised(
            13.6488662517, -0.128744123166, -0.0174837749244, -4.47119633965, tot
        )
    if variant == 3:
        if isinstance(tot, (int, np.integer)):
            return _SLEWS[tot]
        return _slew_tabulated(np.array(_SLEWS), tot)

    raise ValueError("Unknown slew calculation variant '{}'".format(variant))


@nb.jit(nopython=True)
def _slew_parametrised(p0, p1, p2, p3, tot):
    return p0 * np.exp(p1 * np.sqrt(tot) + p2 * tot) + p3


@nb.jit(nopython=True)
def _slew_tabulated(tab, tots):
    n = len(tots)
    out = np.empty(n)
    for i in range(n):
        out[i] = tab[tots[i]]
    return out


_SLEWS = np.array(
    [
        8.01,
        7.52,
        7.05,
        6.59,
        6.15,
        5.74,
        5.33,
        4.95,
        4.58,
        4.22,
        3.89,
        3.56,
        3.25,
        2.95,
        2.66,
        2.39,
        2.12,
        1.87,
        1.63,
        1.40,
        1.19,
        0.98,
        0.78,
        0.60,
        0.41,
        0.24,
        0.07,
        -0.10,
        -0.27,
        -0.43,
        -0.59,
        -0.75,
        -0.91,
        -1.08,
        -1.24,
        -1.41,
        -1.56,
        -1.71,
        -1.85,
        -1.98,
        -2.11,
        -2.23,
        -2.35,
        -2.47,
        -2.58,
        -2.69,
        -2.79,
        -2.89,
        -2.99,
        -3.09,
        -3.19,
        -3.28,
        -3.37,
        -3.46,
        -3.55,
        -3.64,
        -3.72,
        -3.80,
        -3.88,
        -3.96,
        -4.04,
        -4.12,
        -4.20,
        -4.27,
        -4.35,
        -4.42,
        -4.49,
        -4.56,
        -4.63,
        -4.70,
        -4.77,
        -4.84,
        -4.90,
        -4.97,
        -5.03,
        -5.10,
        -5.16,
        -5.22,
        -5.28,
        -5.34,
        -5.40,
        -5.46,
        -5.52,
        -5.58,
        -5.63,
        -5.69,
        -5.74,
        -5.80,
        -5.85,
        -5.91,
        -5.96,
        -6.01,
        -6.06,
        -6.11,
        -6.16,
        -6.21,
        -6.26,
        -6.31,
        -6.36,
        -6.41,
        -6.45,
        -6.50,
        -6.55,
        -6.59,
        -6.64,
        -6.68,
        -6.72,
        -6.77,
        -6.81,
        -6.85,
        -6.89,
        -6.93,
        -6.98,
        -7.02,
        -7.06,
        -7.09,
        -7.13,
        -7.17,
        -7.21,
        -7.25,
        -7.28,
        -7.32,
        -7.36,
        -7.39,
        -7.43,
        -7.46,
        -7.50,
        -7.53,
        -7.57,
        -7.60,
        -7.63,
        -7.66,
        -7.70,
        -7.73,
        -7.76,
        -7.79,
        -7.82,
        -7.85,
        -7.88,
        -7.91,
        -7.94,
        -7.97,
        -7.99,
        -8.02,
        -8.05,
        -8.07,
        -8.10,
        -8.13,
        -8.15,
        -8.18,
        -8.20,
        -8.23,
        -8.25,
        -8.28,
        -8.30,
        -8.32,
        -8.34,
        -8.37,
        -8.39,
        -8.41,
        -8.43,
        -8.45,
        -8.47,
        -8.49,
        -8.51,
        -8.53,
        -8.55,
        -8.57,
        -8.59,
        -8.61,
        -8.62,
        -8.64,
        -8.66,
        -8.67,
        -8.69,
        -8.70,
        -8.72,
        -8.74,
        -8.75,
        -8.76,
        -8.78,
        -8.79,
        -8.81,
        -8.82,
        -8.83,
        -8.84,
        -8.86,
        -8.87,
        -8.88,
        -8.89,
        -8.90,
        -8.92,
        -8.93,
        -8.94,
        -8.95,
        -8.96,
        -8.97,
        -8.98,
        -9.00,
        -9.01,
        -9.02,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
        -9.04,
    ]
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
======================
Basic Analysis Example
======================

"""

# Authors: Tamás Gál <tgal@km3net.de>, Moritz Lotze <mlotze@km3net.de>
# License: BSD-3
# Date: 2017-10-10
# Status: Under construction...
#
# sphinx_gallery_thumbnail_number = 5


#####################################################
# Preparation
# -----------
# The very first thing we do is importing our libraries and setting up
# the Jupyter Notebook environment.

import matplotlib.pyplot as plt  # our plotting module
import pandas as pd  # the main HDF5 reader
import numpy as np  # must have
import km3pipe as kp  # some KM3NeT related helper functions
import seaborn as sns  # beautiful statistical plots!
from km3net_testdata import data_path

#####################################################
# this is just to make our plots a bit "nicer", you can skip it
import km3pipe.style

km3pipe.style.use("km3pipe")

#####################################################
# Accessing the Data File(s)
# --------------------------
# In the following, we will work with one random simulation file with
# reconstruction information from JGandalf which has been converted
# from ROOT to HDF5 using the ``h5extract`` command line tool provided by
# ``KM3Pipe``.
#
# You can find the documentation here:
# https://km3py.pages.km3net.de/km3pipe/cmd.html#h5extract

#####################################################
# Note for Lyon Users
# ~~~~~~~~~~~~~~~~~~~
# If you are working on the Lyon cluster, you just need to load the
# Python module with ``module load python`` and you are all set.

#####################################################
# Converting from ROOT to HDF5 (if needed)
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# Choose a file (take e.g. one from /in2p3/km3net/mc/...),
# load the appropriate Jpp/Aanet version and convert it via::
#
#     h5extract /path/to/a/reconstructed/file.root
#
# You can toggle a few options to include or exclude specific information.
# By default, everything will be extracted but you might want to skip
# Example the hit information. Have a look at ``h5extract -h``.
#
# You might also just pick some of the already converted files from
# HPSS/iRODS!

#####################################################
# First Look at the Data
# ----------------------

filepath = data_path("hdf5/basic_analysis_sample.h5")

#####################################################
# We can have a quick look at the file with the ``ptdump`` command
# in the terminal::
#
#     ptdump filename.h5
#
# For further information, check out the documentation of the KM3NeT HDF5
# format definition: http://km3pipe.readthedocs.io/en/latest/hdf5.html
#

#####################################################
# The ``/event_info`` table contains general information about each event.
# The data is a simple 2D table and each event is represented by a single row.
#
# Let's have a look at the first few rows:
event_info = pd.read_hdf(filepath, "/event_info")
print(event_info.head(5))

#####################################################
# You can easily inspect the columns/fields of a ``Pandas.Dataframe`` with
# the ``.dtypes`` attribute:

print(event_info.dtypes)

#####################################################
# And access the data either by the property syntax (if it's a valid Python
# identifier) or the dictionary syntax, for example to access the neutrino
# weights:

print(event_info.weight_w2)  # property syntax
print(event_info["weight_w2"])  # dictionary syntax


#####################################################
# Next, we will read out the MC tracks which are stored under ``/mc_tracks``.

tracks = pd.read_hdf(filepath, "/mc_tracks")

#####################################################
# It has a similar structure, but now you can have multiple rows which belong
# to an event. The ``event_id`` column holds the ID of the corresponding event.

print(tracks.head(10))

#####################################################
# We now are accessing the first track for each event by grouping via
# ``event_id`` and calling the ``first()`` method of the
# ``Pandas.DataFrame`` object.

primaries = tracks.groupby("event_id").first()

#####################################################
# Here are the first 5 primaries:
print(primaries.head(5))

#####################################################
# Creating some Fancy Graphs
# --------------------------

#####################################################
#
plt.hist(primaries.energy, bins=100, log=True)
plt.xlabel("energy [GeV]")
plt.ylabel("number of events")
plt.title("Energy Distribution")

#####################################################
#
primaries.bjorkeny.hist(bins=100)
plt.xlabel("bjorken-y")
plt.ylabel("number of events")
plt.title("bjorken-y Distribution")

#####################################################
#
zeniths = kp.math.zenith(primaries.filter(regex="^dir_.?$"))
primaries["zenith"] = zeniths

plt.hist(np.cos(primaries.zenith), bins=21, histtype="step", linewidth=2)
plt.xlabel(r"cos($\theta$)")
plt.ylabel("number of events")
plt.title("Zenith Distribution")

#####################################################
#
# Starting positions of primaries
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
plt.hist2d(primaries.pos_x, primaries.pos_y, bins=100, cmap="viridis")
plt.xlabel("x [m]")
plt.ylabel("y [m]")
plt.title("2D Plane")
plt.colorbar()

#####################################################
#
# If you have seaborn installed (`pip install seaborn`), you can easily create
# nice jointplots:
try:
    import seaborn as sns  # noqa

    km3pipe.style.use("km3pipe")  # reset matplotlib style
except:
    print("No seaborn found, skipping example.")
else:
    g = sns.jointplot(x="pos_x", y="pos_y", data=primaries, kind="hex")
    g.set_axis_labels("x [m]", "y[m]")
    plt.subplots_adjust(right=0.90)  # make room for the colorbar
    plt.title("2D Plane")
    plt.colorbar()
    plt.legend()

#####################################################
#
from mpl_toolkits.mplot3d import Axes3D  # noqa

fig = plt.figure()
ax = fig.add_subplot(111, projection="3d")
ax.scatter3D(primaries.pos_x, primaries.pos_y, primaries.pos_z, s=3)
ax.set_xlabel("x [m]", labelpad=10)
ax.set_ylabel("y [m]", labelpad=10)
ax.set_zlabel("z [m]", labelpad=10)
ax.set_title("3D Plane")

#####################################################
#
gandalfs = pd.read_hdf(filepath, "/reco/gandalf")
print(gandalfs.head(5))

#####################################################
#
gandalfs.columns

#####################################################
#
# plt.hist(gandalfs['lambda'], bins=50, log=True)
# plt.xlabel('lambda parameter')
# plt.ylabel('count')
# plt.title('Lambda Distribution of Reconstructed Events')

#####################################################
#
gandalfs["zenith"] = kp.math.zenith(gandalfs.filter(regex="^dir_.?$"))

plt.hist((gandalfs.zenith - primaries.zenith).dropna(), bins=100)
plt.xlabel(r"true zenith - reconstructed zenith [rad]")
plt.ylabel("count")
plt.title("Zenith Reconstruction Difference")

#####################################################
#
l = 0.2
lambda_cut = gandalfs["lambda"] < l
plt.hist((gandalfs.zenith - primaries.zenith)[lambda_cut].dropna(), bins=100)
plt.xlabel(r"true zenith - reconstructed zenith [rad]")
plt.ylabel("count")
plt.title("Zenith Reconstruction Difference for lambda < {}".format(l))

#####################################################
# Combined zenith reco plot for different lambda cuts
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

fig, ax = plt.subplots()
for l in [100, 5, 2, 1, 0.1]:
    l_cut = gandalfs["lambda"] < l
    ax.hist(
        (primaries.zenith - gandalfs.zenith)[l_cut].dropna(),
        bins=100,
        label=r"$\lambda$ = {}".format(l),
        alpha=0.7,
    )
plt.xlabel(r"true zenith - reconstructed zenith [rad]")
plt.ylabel("count")
plt.legend()
plt.title("Zenith Reconstruction Difference for some Lambda Cuts")

#####################################################
# Fitting Angular resolutions
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# Let's fit some distributions: gaussian + lorentz (aka norm + cauchy)
#
# Fitting the gaussian to the whole range is a very bad fit, so
# we make a second gaussian fit only to +- 10 degree.
# Conversely, the Cauchy (lorentz) distribution is a near perfect fit
# (note that ``2 gamma = FWHM``).

from scipy.stats import cauchy, norm  # noqa

residuals = gandalfs.zenith - primaries.zenith
cut = (gandalfs["lambda"] < l) & (np.abs(residuals) < 2 * np.pi)
residuals = residuals[cut]
event_info[cut]

# convert rad -> deg
residuals = residuals * 180 / np.pi

pi = 180
# x axis for plotting
x = np.linspace(-pi, pi, 1000)

c_loc, c_gamma = cauchy.fit(residuals)
fwhm = 2 * c_gamma

g_mu_bad, g_sigma_bad = norm.fit(residuals)
g_mu, g_sigma = norm.fit(residuals[np.abs(residuals) < 10])

plt.hist(residuals, bins="auto", label="Histogram", density=True, alpha=0.7)
plt.plot(
    x,
    cauchy(c_loc, c_gamma).pdf(x),
    label="Lorentz: FWHM $=${:.3f}".format(fwhm),
    linewidth=2,
)
plt.plot(
    x,
    norm(g_mu_bad, g_sigma_bad).pdf(x),
    label="Unrestricted Gauss: $\sigma =$ {:.3f}".format(g_sigma_bad),
    linewidth=2,
)
plt.plot(
    x,
    norm(g_mu, g_sigma).pdf(x),
    label="+- 10 deg Gauss: $\sigma =$ {:.3f}".format(g_sigma),
    linewidth=2,
)
plt.xlim(-pi / 4, pi / 4)
plt.xlabel("Zenith residuals / deg")
plt.legend()

####################################################################
# We can also look at the median resolution without doing any fits.
#
# In textbooks, this metric is also called Median Absolute Deviation.

resid_median = np.median(residuals)
residuals_shifted_by_median = residuals - resid_median
absolute_deviation = np.abs(residuals_shifted_by_median)
resid_mad = np.median(absolute_deviation)

plt.hist(np.abs(residuals), alpha=0.7, bins="auto", label="Absolute residuals")
plt.axvline(resid_mad, label="MAD: {:.2f}".format(resid_mad), linewidth=3)
plt.title("Average resolution: {:.3f} degree".format(resid_mad))
plt.legend()
plt.xlabel("Absolute zenith residuals / deg")



# Filename: widgets.py
"""
GUI elements for the pipeinspector.

"""

import urwid
import pprint
import math
import sys

from pipeinspector.settings import UI

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class BlobBrowser(urwid.Frame):
    def __init__(self):
        self.items = []
        self.cursor_position = 0

        self.header = urwid.AttrMap(urwid.Text("Keys:"), "head")

        self.listbox = urwid.ListBox(urwid.SimpleListWalker(self.items))
        self.frame = urwid.Frame(self.listbox, header=self.header)
        line_box = urwid.AttrMap(urwid.LineBox(self.frame), "body")
        urwid.Frame.__init__(self, line_box)
        self.overlay = None
        self.popup = None

    def load(self, blob):
        del self.listbox.body[:]
        new_items = []
        for key in sorted(blob.keys()):
            item_widget = ItemWidget(key, blob[key])
            new_items.append(item_widget)
            urwid.connect_signal(item_widget, "key_selected", self.key_selected)
        self.listbox.body.extend(new_items)
        self.listbox.set_focus(self.cursor_position)

    def key_selected(self, data):
        def formatter(obj):
            if hasattr(obj, "__insp__"):
                return obj.__insp__()
            if hasattr(obj, "size"):
                output = ""
                for obj in data:
                    output += str(obj) + "\n"
                return output
            return pprint.pformat(obj)

        content = [urwid.Text(line) for line in formatter(data).split("\n")]
        self.popup = urwid.ListBox(content)
        popup_box = urwid.LineBox(self.popup)
        self.overlay = urwid.Overlay(
            popup_box,
            self.body,
            "center",
            ("relative", 80),
            "middle",
            ("relative", 80),
        )
        self.body = self.overlay

    def keypress(self, size, key):
        input = urwid.Frame.keypress(self, size, key)
        self.cursor_position = self.listbox.focus_position
        if self.overlay:
            if input in UI.keys["escape"]:
                self.body = self.overlay.bottom_w
                self.overlay = None
        else:
            return input


class ItemWidget(urwid.WidgetWrap):
    signals = ["key_selected"]

    def __init__(self, key, data):
        self.key = key
        self.data = data
        try:
            size_label = str(len(data)) + " items"
        except TypeError:
            size_label = str(sys.getsizeof(data)) + " bytes"

        type_label = str(type(data)).split("'")[1]

        self.item = [
            (
                "fixed",
                35,
                urwid.Padding(urwid.AttrWrap(urwid.Text(key), "body", "focus"), left=2),
            ),
            urwid.AttrWrap(urwid.Text(type_label), "body", "focus"),
            urwid.AttrWrap(urwid.Text(size_label), "body", "focus"),
        ]
        w = urwid.Columns(self.item)
        self.__super.__init__(w)

    def selectable(self):
        return True

    def keypress(self, size, key):
        if key == "x":
            urwid.emit_signal(self, "key_selected", self.data)
        return key


class BlobWidget(urwid.Pile):
    signals = ["blob_selected"]

    def __init__(self):
        self.width = 50
        self.size = (0,)
        self.index = 0
        urwid.Pile.__init__(
            self,
            [
                urwid.Text("", wrap="clip"),
                urwid.Text("", wrap="clip"),
                urwid.Text("", wrap="clip"),
            ],
        )

    def goto_blob(self, position):
        self.index = position
        self._emit_blob_selected()
        self.draw()

    def previous_blob(self, step=1):
        self.index -= step
        if self.index <= 0:
            self.index = 0
        self._emit_blob_selected()
        self.draw()

    def next_blob(self, step=1):
        self.index += step
        self._emit_blob_selected()
        self.draw()

    def draw(self):
        self.widget_list[0].set_text(self._make_blob_icons(self.index))
        self.widget_list[1].set_text(self._make_ruler(self.index))
        self.widget_list[2].set_text(self._make_scale_labels(self.index))

    def render(self, size, focus):
        self.size = size
        self.draw()
        return urwid.Pile.render(self, size, focus)

    def _emit_blob_selected(self):
        urwid.emit_signal(self, "blob_selected", self.index)

    def _make_ruler(self, start):
        if start <= 10:
            start = 0
        else:
            start -= 10
        segment = "|    '    "
        repeat = int(math.ceil(self.width / len(segment)) + 1)
        ruler = segment * repeat
        slice_start = start % len(segment)
        slice_end = (start % len(segment)) + self.width
        return ruler[slice_start:slice_end]

    def _make_scale_labels(self, start):
        if start <= 10:
            start = 0
        else:
            start -= 10
        lowest_tick = int(math.floor(start / 10) * 10)
        highest_tick = lowest_tick + self.width
        ticks_labels = [
            "{0:<10}".format(i) for i in range(lowest_tick, highest_tick, 10)
        ]
        slice_start = start % 10
        slice_end = (start % 10) + self.width
        ticks = "".join(ticks_labels)[slice_start:slice_end]
        return ticks

    def _make_blob_icons(self, start):
        icon = "B"
        if start < 10:
            icons = "." + icon * (self.width - 1)
        else:
            icons = icon * self.width
        if start > 10:
            start = 10
        return [
            ("blob", icons[:start]),
            ("blob_selected", icons[start]),
            ("blob", icons[start + 1 :]),
        ]



"""
Print the structure of a hdf5 file.

This is a less verbose version of ``ptdump`` from the pytables package.
If you want much more detailed (+verbose) output, e.g. for debugging, by
all means use the ``ptdump`` util (it's already installed as a dependency
alongside km3pipe).

Usage:
    h5tree [options] FILE
    h5tree (-h | --help)
    h5tree --version

Options:
    FILE        Input file.
    -h --help   Show this screen.
    --no-meta   Don't print meta data at top
    --titles    Print leaf titles.
    --classic   Print like ptdump.

"""

import numpy as np
import tables

__author__ = "Moritz Lotze"
__copyright__ = "Copyright 2016, Moritz Lotze and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Moritz Lotze"
__email__ = "mlotze@km3net.de"
__status__ = "Development"


def nodeinfo(node, print_titles=False):
    pathname = node._v_pathname
    try:
        n_rows = node.shape[0]
        n_rows_str = "{}".format(n_rows)
    except AttributeError:
        n_rows_str = None
    try:
        title = str(node.title)
        if not print_titles or not title or title.isspace():
            title = None
        else:
            title = "'{}'".format(title)
    except AttributeError:
        title = None
    return ", ".join(
        [node for node in (pathname, title, n_rows_str) if node is not None]
    )


def meta(h5):
    try:
        version = np.bytes_(h5.root._v_attrs.format_version)
        print("KM3HDF5 v{}".format(version.decode("utf-8")))
    except AttributeError:
        pass
    try:
        info = h5.root.event_info
        n_events = info.shape[0]
        print("Number of Events: {}".format(n_events))
    except tables.NoSuchNodeError:
        pass


def h5tree(h5name, print_meta=True, pretty=True, **kwargs):
    with tables.open_file(h5name) as h5:
        if print_meta:
            meta(h5)
        node_kinds = h5.root._v_file._node_kinds[1:]
        what = h5.root._f_walk_groups()
        for k, group in enumerate(what):
            d = group._v_depth
            if pretty:
                print("│  " * (d - 1) + "├── " * d + group._v_name)
            else:
                print(group._v_pathname)
            for kind in node_kinds:
                for n, node in enumerate(group._f_list_nodes(kind)):
                    if n == 0:
                        n_siblings = node._v_parent._v_nchildren
                    filler = "├── "
                    if n == n_siblings - 1:
                        filler = "└── "
                    d = node._v_depth
                    if pretty:
                        print("│  " * (d - 1) + filler + node._v_name)
                    else:
                        print(nodeinfo(node, **kwargs))


def main():
    from docopt import docopt

    args = docopt(__doc__)
    fname = args["FILE"]
    do_titles = bool(args["--titles"])
    do_meta = not bool(args["--no-meta"])
    do_classic = bool(args["--classic"])
    h5tree(
        fname,
        print_titles=do_titles,
        print_meta=do_meta,
        print_pretty=not (do_classic),
    )



"""
A tool to extract data from KM3NeT ROOT files (offline format)
to HDF5 fast (hence the f).

Usage:
    h5extractf [options] FILENAME
    h5extractf (-h | --help)

Options:
    -o OUTFILE                  Output file.
    --provenance-file=FILENAME  The file to store the provenance information.
    --without-full-reco         Don't include all reco tracks, only the best.
    --without-calibration       Don't include calibration information for offline hits.
    -h --help                   Show this screen.

"""

import time
from uuid import uuid4
import numpy as np
import h5py
import awkward as ak
from thepipe import Provenance
import km3io
import km3pipe as kp

FORMAT_VERSION = np.bytes_("5.1")

log = kp.logger.get_logger("h5extractf")


def h5extractf(
    root_file, outfile=None, without_full_reco=False, without_calibration=False
):
    if without_calibration:
        calibration_fields = []
    else:
        calibration_fields = [
            "pos_x",
            "pos_y",
            "pos_z",
            "dir_x",
            "dir_y",
            "dir_z",
            "tdc",
        ]
    fields = {
        "event_info": [
            ("id", "event_id"),  # id gets renamed to event_id
            "run_id",
            ("t_sec", "timestamp"),
            ("t_ns", "nanoseconds"),
            ("mc_t", "mc_time"),
            "trigger_mask",
            "trigger_counter",
            "overlays",
            "det_id",
            "frame_index",
            "mc_run_id",
        ],
        # weights get put into event_info as well
        "event_info_weights": [
            "weight_w1",
            "weight_w2",
            "weight_w3",
            "weight_w4",
        ],
        "hits": [
            "channel_id",
            "dom_id",
            ("t", "time"),
            "tot",
            ("trig", "triggered"),
            *calibration_fields,
        ],
        "mc_hits": [
            "a",
            "origin",
            "pmt_id",
            ("t", "time"),
        ],
        "tracks": [
            "pos_x",
            "pos_y",
            "pos_z",
            "dir_x",
            "dir_y",
            "dir_z",
            "E",
            "t",
            ("len", "length"),
            "rec_type",
            ("lik", "likelihood"),
            "id",
        ],
        "mc_tracks": [
            "pos_x",
            "pos_y",
            "pos_z",
            "dir_x",
            "dir_y",
            "dir_z",
            ("E", "energy"),
            ("t", "time"),
            ("len", "length"),
            "pdgid",
            "id",
        ],
    }

    if outfile is None:
        outfile = root_file + ".h5"

    start_time = time.time()
    with h5py.File(outfile, "w") as f:

        _uuid = str(uuid4())
        Provenance().record_output(outfile, uuid=_uuid, comment="Converted HDF5 file")

        with km3io.OfflineReader(root_file) as r:
            Provenance().record_input(
                root_file, uuid=str(r.uuid), comment="Input ROOT file"
            )

            if r.header is not None:
                print("Processing header")
                f.create_dataset(
                    "raw_header",
                    data=kp.io.hdf5.header2table(r.header),
                )
            print("Processing event_info")
            np_event_info = _branch_to_numpy(r, fields["event_info"])
            np_weights = _ak_to_numpy(r.w, fields["event_info_weights"])
            np_event_info[0].update(np_weights[0])
            np_w2 = _parse_w2list(r)
            if np_w2 is not None:
                np_event_info[0].update(np_w2[0])
            _to_hdf(f, "event_info", np_event_info)

            # TODO remove group_info once km3pipe does not require it anymore
            group_info = np.core.records.fromarrays(
                [np.arange(len(np_event_info[1]))], names=["group_id"]
            )
            f.create_dataset("group_info", data=group_info)

            print("Processing tracks")
            reco = f.create_group("reco")
            for branch_data in _yield_tracks(
                r.tracks, fields["tracks"], without_full_reco=without_full_reco
            ):
                _to_hdf(reco, *branch_data)

            for field_name in ("hits", "mc_hits", "mc_tracks"):
                if r[field_name] is None:
                    continue
                print("Processing", field_name)
                np_branch = _branch_to_numpy(r[field_name], fields[field_name])
                if np_branch[1].sum() == 0:
                    # empty branch, e.g. mc_hits for data files
                    continue
                _to_hdf(f, field_name, np_branch)
        f.attrs.create("format_version", FORMAT_VERSION)
        f.attrs.create("km3pipe", kp.__version__)
        f.attrs.create("origin", root_file)
        f.attrs.create("kid", _uuid)
    print("Completed in {:.1f} s".format(time.time() - start_time))


def _parse_w2list(f):
    try:
        w2list_idx = km3io.tools.get_w2list_idx(f)
    except AttributeError:
        return None
    if w2list_idx is None:
        return None
    return _ak_to_numpy(f.w2list, list(w2list_idx.values()))


def _branch_to_numpy(branch, fields):
    """
    Read 1D or 2D fields from a branch and convert them to numpy arrays.

    Parameters
    ----------
    branch : km3io.rootio.Branch
        Branch of an offline file.
    fields : list
        The field names. If an entry is a tuple instead of a str,
        first entry is the field name, secound entry is its h5 name.

    Returns
    -------
    tuple
        Numpy-fied awkward array, length 2.
        First entry is a dict with fields as keys, and 1D np.arrays as values.
        Secound entry is a 1D np.array, the number of items for each event.

    """
    data, n_items = {}, None
    # would be better to read out all fields at once
    for field in fields:
        if isinstance(field, (tuple, list)):
            field, column = field
        else:
            column = field
        d = branch.__getattr__(field)
        n_dims = d.ndim
        if n_dims == 1:
            if n_items is None:
                n_items = np.ones(len(d), dtype="int64")
        elif n_dims == 2:
            if n_items is None:
                n_items = ak.num(d).to_numpy()
            d = ak.flatten(d)
        else:
            raise ValueError("Can not process field", field)
        data[column] = d.to_numpy()
    return data, n_items


def _ak_to_numpy(ak_array, fields):
    """
    Convert the given awkward array to a numpy table.

    Parameters
    ----------
    ak_array : awkward.Array
        The awkward array, 2D or 3D.
    fields : List
        The column names of the last axis of the array.

    Returns
    -------
    np_branch : tuple
        Numpy-fied awkward array. See output of _branch_to_numpy.

    """
    n_dims = ak_array.ndim - 1
    if n_dims == 1:
        n_items = np.ones(len(ak_array), dtype="int64")
    elif n_dims == 2:
        n_items = ak.num(ak_array).to_numpy()
        ak_array = ak.flatten(ak_array)
    else:
        raise ValueError("Cannot process array")
    if len(ak_array) == 0:
        return {
            fields[i]: ak_array.to_numpy().reshape((0,)) for i in range(len(fields))
        }, n_items

    filled = np.ma.filled(
        ak.pad_none(ak_array, target=len(fields), axis=-1).to_numpy(),
        fill_value=np.nan,
    )
    return {fields[i]: filled[:, i] for i in range(len(fields))}, n_items


def _yield_tracks(tracks, fields, without_full_reco=False):
    """
    Yield info from the tracks branch.

    Parameters
    ----------
    tracks : km3io.rootio.Branch
        The tracks branch.
    fields : list
        The fields to read for each track.
    without_full_reco : bool
        Don't include all reco tracks, only the best.

    Yields
    ------
    name : str
        Name of how the tracks where chosen.
    np_branch : tuple
        Numpy-fied awkward array. See output of _branch_to_numpy.

    """

    nmbr_events = len(tracks)
    if nmbr_events == 0:
        log.warn("No events found")
        return

    track_fmap = {
        "best_jmuon": (
            km3io.definitions.reconstruction.JMUONPREFIT,
            km3io.tools.best_jmuon,
        ),
        "best_jshower": (
            km3io.definitions.reconstruction.JSHOWERPREFIT,
            km3io.tools.best_jshower,
        ),
        "best_dusjshower": (
            km3io.definitions.reconstruction.DUSJSHOWERPREFIT,
            km3io.tools.best_dusjshower,
        ),
        "best_aashower": (
            km3io.definitions.reconstruction.AASHOWERFITPREFIT,
            km3io.tools.best_aashower,
        ),
    }
    if not without_full_reco:
        track_fmap["tracks"] = (None, None)

    n_columns = max(km3io.definitions.fitparameters.values()) + 1

    for name, (stage, func) in track_fmap.items():
        if (stage is None) or (ak.any(stage == tracks.rec_stages)):
            if func is None:
                sel_tracks = tracks
            else:
                sel_tracks = func(tracks)
            np_branch = _branch_to_numpy(sel_tracks, fields)

            try:
                np_fitinf = _ak_to_numpy(sel_tracks.fitinf, range(n_columns))
            except ValueError as e:
                log.warning(
                    "Cannot convert fitinf (%s): probably unsupported KM3NeT dataformat version or empty branch. Skipping..."
                    % e
                )
                continue
            for fitparam, idx in km3io.definitions.fitparameters.items():
                np_branch[0][fitparam] = np_fitinf[0][idx].astype("float32")

            yield name, np_branch


def _to_hdf(f, name, np_branch, with_group_id=True):
    """
    Save a numpy-fied awkward array as a dataset to hdf5.

    Parameters
    ----------
    f : h5py.File
        The opened h5 file.
    name : str
        Name of the dataset to create.
    np_branch : tuple
        The data to save. See output of _branch_to_numpy.
    with_group_id : bool
        If False (default), save with indices as a sperate dataset.
        If true, save with group_id as an additional column.

    """
    # TODO set default of with_group_id to False once km3pipe supports
    #  indexed recarrays
    hdf_settings = {
        "compression": "gzip",
        "compression_opts": 5,
        "shuffle": True,
        "fletcher32": True,
        "chunks": True,
    }
    data, n_items = np_branch
    if with_group_id:
        data["group_id"] = np.repeat(np.arange(len(n_items)), n_items)
    else:
        index = np.concatenate([[0], np.cumsum(n_items)[:-1]])
        indices = np.core.records.fromarrays(
            [index, n_items], names=["index", "n_items"]
        )
        f.create_dataset(name + "_indices", data=indices, **hdf_settings)

    f.create_dataset(
        name,
        data=np.core.records.fromarrays(list(data.values()), names=list(data.keys())),
        **hdf_settings,
    )


def main():
    from docopt import docopt

    args = docopt(__doc__, version=kp.version)

    outfile = args["-o"]
    if outfile is None:
        outfile = args["FILENAME"] + ".h5"

    provfile = args["--provenance-file"]
    if provfile is None:
        provfile = outfile + ".prov.json"

    Provenance().outfile = provfile

    h5extractf(
        args["FILENAME"],
        outfile,
        without_full_reco=args["--without-full-reco"],
        without_calibration=args["--without-calibration"],
    )


if __name__ == "__main__":
    main()



# Filename: communication.py
# -*- coding: utf-8 -*-
# pylint: disable=locally-disabled
"""
Talk to the ELOG, interact with the Chat server etc.

"""

import builtins
import os
import passlib
import requests
import time

import km3pipe as kp


class ELOGService(kp.Module):
    """Allows communication with the KM3NeT ELOG server.

    Parameters
    ----------
    username: string (optional)
    password: string (required)
    url: string (optional)

    """

    def configure(self):
        self.url = self.get("url", default="https://elog.km3net.de/")
        self._password = create_elog_password_hash(self.require("password"))
        self._username = self.get("username", default="km3net")
        self.expose(self.post_elog, "post_elog")

    def post_elog(
        self, logbook, subject, message, author, message_type="Comment", files=None
    ):
        """Post an ELOG entry

        Make sure to provide an existing `author`.

        Parameters
        ----------
        logbook: string (e.g. "individual+Logbooks")
        subject: string
        message: string
        author: string (make sure to provide an existing one!)
        message_type: string (default="Comment")
        files: list of strings (filenames, optional)
        """
        data = {
            "exp": logbook,
            "cmd": "Submit",
            "Subject": subject,
            "Author": author,
            "Type": message_type,
            "unm": self._username,
            "upwd": self._password,
            "When": int(time.time()),
        }
        if files is None:
            files = []
        files = [
            ("attfile{}".format(idx + 1), (os.path.basename(f), builtins.open(f, "rb")))
            for idx, f in enumerate(files)
        ]
        files.append(("Text", ("", message)))
        try:
            r = requests.post(
                self.url + "/" + logbook,
                data=data,
                files=files,
                allow_redirects=False,
                verify=False,
            )
        except requests.RequestException as e:
            self.log.error("Cannot reach the ELOG server!\nError: {}".format(e))
        else:
            if r.status_code not in [200, 302]:
                self.log.error(
                    "Something went wrong...\n\nHere is what we got:\n{}".format(
                        r.content.decode("utf-8", "ignore")
                    )
                )
            else:
                self.cprint("ELOG post created successfully.")

        for (_, (_, fobj)) in files:
            if fobj and hasattr(fobj, "close"):
                fobj.close()


def create_elog_password_hash(password):
    """Create a SHA256 encrypted password for ELOGs."""
    from passlib.hash import sha256_crypt

    return sha256_crypt.encrypt(password, salt="", rounds=5000)[4:]



# Filename: hardware.py
# -*- coding: utf-8 -*-
# pylint: disable=locally-disabled
"""
A collection of controllers and hardware related stuff.

"""

import time
import os

import km3pipe as kp

__author__ = "Jonas Reubelt and Tamas Gal"
__email__ = "jreubelt@km3net.de"
__status__ = "Development"

log = kp.logger.get_logger(__name__)  # pylint: disable=C0103


class PhidgetsController(kp.Module):
    def configure(self):
        from Phidgets.Devices.Stepper import Stepper
        from Phidgets.Devices.Encoder import Encoder

        self.current_limit = self.get("current_limit") or 2.5
        self.motor_id = self.get("motor_id") or 0
        self.stepper = Stepper()
        self.encoder = Encoder()
        self.setup(self.motor_id)

    def setup(self, motor_id):
        self.stepper.openPhidget()
        self.stepper.waitForAttach(10000)

        self.encoder.openPhidget()
        self.encoder.waitForAttach(10000)

        self.stepper.setVelocityLimit(motor_id, 1000)
        self.stepper.setAcceleration(motor_id, 5000)
        self.stepper.setCurrentLimit(motor_id, 2.5)

        self.e = 13250.0
        self.s = 70500.0

        self._stepper_dest = 0
        self._encoder_dest = 0

        self.reset_positions()

    def drive_to_angle(self, ang, motor_id=0, relative=False):
        stepper_dest = self._stepper_dest = self.raw_stepper_position(ang)
        self._encoder_dest = self.raw_encoder_position(ang)

        if relative:
            self.reset_positions()

        self.wake_up()
        time.sleep(0.1)

        self.stepper_target_pos = stepper_dest
        self.wait_for_stepper()
        self.log_offset()

        while abs(self.offset) > 1:
            self.log_offset()
            stepper_offset = round(self.offset / self.e * self.s)
            log.debug("Correcting stepper by {0}".format(stepper_offset))
            log.debug("Stepper target pos: {0}".format(self.stepper_target_pos))
            log.debug("Stepper pos: {0}".format(self.stepper_pos))
            self.stepper_target_pos = self.stepper_pos + stepper_offset
            self.wait_for_stepper()

        self.log_positions()
        self.stand_by()

    def wait_for_stepper(self):
        while self.stepper_pos != self._stepper_dest:
            time.sleep(0.1)
            self.log_positions()

    def log_offset(self):
        log.debug("Difference (encoder): {0}".format(self.offset))

    @property
    def offset(self):
        return self._encoder_dest - self.encoder_pos

    @property
    def stepper_target_pos(self):
        return self.stepper.getTargetPosition(self.motor_id)

    @stepper_target_pos.setter
    def stepper_target_pos(self, val):
        self._stepper_dest = int(val)
        self.stepper.setTargetPosition(self.motor_id, int(val))

    @property
    def stepper_pos(self):
        return self.stepper.getCurrentPosition(self.motor_id)

    @stepper_pos.setter
    def stepper_pos(self, val):
        self.stepper.setCurrentPosition(self.motor_id, int(val))

    @property
    def encoder_pos(self):
        return self.encoder.getPosition(self.motor_id)

    def raw_stepper_position(self, angle):
        return round(angle * self.s / 360)

    def raw_encoder_position(self, angle):
        return round(angle * self.e / 360)

    def wake_up(self, motor_id=0):
        self.stepper.setEngaged(motor_id, 1)

    def stand_by(self, motor_id=0):
        self.stepper.setEngaged(motor_id, 0)

    def reset_positions(self, motor_id=0):
        self.stepper.setCurrentPosition(motor_id, 0)
        self.encoder.setPosition(motor_id, 0)

    def log_positions(self, motor_id=0):
        log.info(
            "Stepper position: {0}\nEncoder position:{1}".format(
                self.stepper_pos / self.s * 360, self.encoder_pos / self.e * 360
            )
        )


class USBTMC(object):
    "USB TMC communicator"

    def __init__(self, path):
        self.device = os.open(path, os.O_RDWR)

    def write(self, msg):
        os.write(self.device, msg)

    def read(self, size=1000):
        return os.read(self.device, size)

    @property
    def name(self):
        self.write(b"*IDN?")
        return self.read()

    def reset(self):
        self.write(b"*RST")


class Agilent33220A(object):
    """Controller for the Arbitrary Waveform Generator"""

    def __init__(self, path):
        self.tmc = USBTMC(path)
        self._output = False
        self._amplitude = None
        self._frequency = None
        self._mode = None

    @property
    def output(self):
        return self._output

    @output.setter
    def output(self, value):
        self.tmc.write("OUTP {0}".format("ON" if value else "OFF").encode())
        self._output = value

    @property
    def amplitude(self):
        return self._amplitude

    @amplitude.setter
    def amplitude(self, val):
        low, high = val
        diff = high - low
        offset = diff / 2
        self.tmc.write("VOLT:OFFS {0}".format(offset).encode())
        self.tmc.write("VOLT {0}".format(diff).encode())
        self._amplitude = val
        self._mode = None

    @property
    def frequency(self):
        return self._frequency

    @frequency.setter
    def frequency(self, val):
        self.tmc.write("FREQ {0}".format(val).encode())
        self._frequency = val

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, val):
        valid_modes = ("sin", "squ", "ramp", "puls", "nois", "dc", "user")
        if val not in valid_modes:
            print(
                "Not a valid mode: '{0}'. Valid modes are: {1}".format(val, valid_modes)
            )
            return
        self.tmc.write("FUNC {0}".format(val.upper()).encode())
        self._mode = val



# Filename: k40.py
# -*- coding: utf-8 -*-
# pylint: disable=locally-disabled
"""
A collection of k40 related functions and modules.

"""

import os
from itertools import combinations
from collections import defaultdict
from functools import partial
from datetime import datetime
import io

import numpy as np
import pickle

import km3pipe as kp
import km3pipe.extras
from km3pipe.io.daq import TMCHData

log = kp.logger.get_logger(__name__)  # pylint: disable=C0103
try:
    from numba import jit
except ImportError:
    log.warning(
        "This module requires `numba` to be installed, otherwise "
        "the functions and Modules imported from this module can "
        "be painfully slow."
    )
    jit = lambda f: f

__author__ = "Jonas Reubelt"
__email__ = "jreubelt@km3net.de"
__status__ = "Development"

# log.setLevel(logging.DEBUG)

TIMESLICE_LENGTH = 0.1  # [s]
MC_ANG_DIST = np.array([-0.72337394, 2.59196335, -0.43594182, 1.10514914])


class K40BackgroundSubtractor(kp.Module):
    """Subtracts random coincidence background from K40 data

    Notes
    -----
    Requires servce 'MedianPMTRates()'
    Writes 'K40Counts' into the Blob: dict, Corrected K40 counts

    """

    def configure(self):
        self.combs = list(combinations(range(31), 2))
        self.mode = self.get("mode", default="online")
        self.expose(self.get_corrected_counts, "GetCorrectedTwofoldCounts")
        self.corrected_counts = None

    def process(self, blob):
        if self.mode != "online":
            return blob
        self.cprint("Subtracting random background calculated from single rates")
        corrected_counts = self.subtract_background()
        blob["CorrectedTwofoldCounts"] = corrected_counts

        return blob

    def get_corrected_counts(self):
        return self.corrected_counts

    def subtract_background(self):
        counts = self.services["TwofoldCounts"]
        dom_ids = list(counts.keys())
        mean_rates = self.services["GetMedianPMTRates"]()
        corrected_counts = {}
        livetimes = self.services["GetLivetime"]()
        for dom_id in dom_ids:
            try:
                pmt_rates = mean_rates[dom_id]
            except KeyError:
                log.warning("Skipping BG correction for DOM {}.".format(dom_id))
                corrected_counts[dom_id] = counts[dom_id]
                continue
            livetime = livetimes[dom_id]
            k40_rates = counts[dom_id] / livetime
            bg_rates = []
            for c in self.combs:
                bg_rates.append(pmt_rates[c[0]] * pmt_rates[c[1]] * 1e-9)
            corrected_counts[dom_id] = (k40_rates.T - np.array(bg_rates)).T * livetime
        return corrected_counts

    def finish(self):
        if self.mode == "offline":
            self.cprint("Subtracting background calculated from summaryslices.")
            self.corrected_counts = self.subtract_background()

    def dump(self, mean_rates, corrected_counts, livetime):
        pickle.dump(mean_rates, open("mean_rates.p", "wb"))
        pickle.dump(
            {"data": corrected_counts, "livetime": livetime},
            open("k40_counts_bg_sub.p", "wb"),
        )


class IntraDOMCalibrator(kp.Module):
    """Intra DOM calibrator which performs the calibration from K40Counts.

    Parameters
    ----------
    det_id: int
      Detector ID [default: 14]
    ctmin: float
      Minimum cos(angle)
    mode: str ('offline' | 'online')
      Calibration mode [default: 'online']

    Notes
    -----
    Requires 'TwofoldCounts': dict
        (key=dom_id, value=matrix of k40 counts 465x(dt*2+1))
    Requires 'CorrectedTwofoldCounts': dict
        (key=dom_id, value=matrix of k40 counts 465x(dt*2+1))
    Writes 'IntraDOMCalibration' into the blob: dict (key=dom_id, value=calibration)

    """

    def configure(self):
        det_id = self.get("det_id") or 14
        self.detector = kp.hardware.Detector(det_id=det_id)
        self.ctmin = self.require("ctmin")
        self.mode = self.get("mode", default="online")
        self.calib_filename = self.get("calib_filename", default="k40_cal.p")

    def process(self, blob):
        if self.mode != "online":
            return blob

        if "CorrectedTwofoldCounts" in blob:
            log.info("Using corrected twofold counts")
            fit_background = False
            twofold_counts = blob["CorrectedTwofoldCounts"]
        else:
            log.info("No corrected twofold counts found, fitting background.")
            twofold_counts = self.services["TwofoldCounts"]
            fit_background = True

        blob["IntraDOMCalibration"] = self.calibrate(twofold_counts, fit_background)
        return blob

    def calibrate(self, twofold_counts, fit_background=False):
        self.cprint("Starting calibration:")
        calibration = {}

        for dom_id, data in twofold_counts.items():
            self.cprint(" calibrating DOM '{0}'".format(dom_id))
            try:
                livetime = self.services["GetLivetime"]()[dom_id]
                calib = calibrate_dom(
                    dom_id,
                    data,
                    self.detector,
                    livetime=livetime,
                    fit_background=fit_background,
                    ad_fit_shape="exp",
                    ctmin=self.ctmin,
                )
            except RuntimeError:
                log.error(" skipping DOM '{0}'.".format(dom_id))
            except KeyError:
                log.error(" skipping DOM '{0}', no livetime".format(dom_id))
            else:
                calibration[dom_id] = calib

        return calibration

    def finish(self):
        if self.mode == "offline":
            self.cprint("Starting offline calibration")
            if "GetCorrectedTwofoldCounts" in self.services:
                self.cprint("Using corrected twofold counts")
                twofold_counts = self.services["GetCorrectedTwofoldCounts"]()
                fit_background = False
            else:
                self.cprint("Using uncorrected twofold counts")
                twofold_counts = self.services["TwofoldCounts"]
                fit_background = True
            calibration = self.calibrate(twofold_counts, fit_background=fit_background)
            self.cprint("Dumping calibration to '{}'.".format(self.calib_filename))
            with open(self.calib_filename, "wb") as f:
                pickle.dump(calibration, f)


class TwofoldCounter(kp.Module):
    """Counts twofold coincidences in timeslice hits per PMT combination.

    Parameters
    ----------
    'tmax': int
      time window of twofold coincidences [ns]
    'dump_filename': str
      name for the dump file

    Notes
    -----
    Requires the key 'TSHits': RawHitSeries
    Provides the following services:
    'TwofoldCounts': dict (key=dom_id, value=matrix (465,(dt*2+1)))
    'ResetTwofoldCounts': reset the TwofoldCounts dict
    'GetLivetime()': dict (key=dom_id, value=float)
    'DumpTwofoldCounts': Writes twofold counts into 'dump_filename'

    """

    def configure(self):
        self.tmax = self.get("tmax") or 20
        self.dump_filename = self.get("dump_filename")

        self.counts = None
        self.n_timeslices = None
        self.start_time = datetime.utcnow()
        self.reset()

        self.expose(self.counts, "TwofoldCounts")
        self.expose(self.get_livetime, "GetLivetime")
        self.expose(self.reset, "ResetTwofoldCounts")
        if self.dump_filename is not None:
            self.expose(self.dump, "DumpTwofoldCounts")

        if "GetSkippedFrames" in self.services:
            self.skipped_frames = self.services["GetSkippedFrames"]()
        else:
            self.skipped_frames = None

    def reset(self):
        """Reset coincidence counter"""
        self.counts = defaultdict(partial(np.zeros, (465, self.tmax * 2 + 1)))
        self.n_timeslices = defaultdict(int)

    def get_livetime(self):
        return {dom_id: n * TIMESLICE_LENGTH for dom_id, n in self.n_timeslices.items()}

    def process(self, blob):
        log.debug("Processing timeslice")
        hits = blob["TSHits"]

        frame_index = blob["TimesliceInfo"].frame_index
        dom_ids = set(np.unique(hits.dom_id))
        if self.skipped_frames is not None:
            skipped_dom_ids = set(self.skipped_frames[frame_index])
        else:
            skipped_dom_ids = set()

        for dom_id in dom_ids - skipped_dom_ids:
            self.n_timeslices[dom_id] += 1
            mask = hits.dom_id == dom_id
            times = hits.time[mask]
            channel_ids = hits.channel_id[mask]
            sort_idc = np.argsort(times, kind="quicksort")
            add_to_twofold_matrix(
                times[sort_idc],
                channel_ids[sort_idc],
                self.counts[dom_id],
                tmax=self.tmax,
            )
        return blob

    def dump(self):
        """Write coincidence counts into a Python pickle"""
        self.cprint("Dumping data to {}".format(self.dump_filename))
        pickle.dump(
            {"data": self.counts, "livetime": self.get_livetime()},
            open(self.dump_filename, "wb"),
        )


class HRVFIFOTimesliceFilter(kp.Module):
    """Creat a frame index lookup table which holds DOM IDs of frames with
    at least one PMT in HRV."""

    def configure(self):
        filename = self.require("filename")
        filter_hrv = self.get("filter_hrv", default=False)
        self.expose(self.get_skipped_frames, "GetSkippedFrames")
        self.skipped_frames = defaultdict(list)
        p = kp.io.jpp.SummaryslicePump(filename=filename)
        for b in p:
            sum_info = b["SummarysliceInfo"]
            frame_index = sum_info.frame_index
            summaryslice = b["Summaryslice"]
            for dom_id, sf in summaryslice.items():
                if not sf["fifo_status"] or (filter_hrv and any(sf["hrvs"])):
                    self.skipped_frames[frame_index].append(dom_id)

    def get_skipped_frames(self):
        return self.skipped_frames


class SummaryMedianPMTRateService(kp.Module):
    def configure(self):
        self.expose(self.get_median_rates, "GetMedianPMTRates")
        self.filename = self.require("filename")

    def get_median_rates(self):
        rates = defaultdict(list)

        if "GetSkippedFrames" in self.services:
            skipped_frames = self.services["GetSkippedFrames"]()
        else:
            skipped_frames = None

        p = kp.io.jpp.SummaryslicePump(filename=self.filename)
        for b in p:
            sum_info = b["SummarysliceInfo"]
            frame_index = sum_info.frame_index
            summary = b["Summaryslice"]
            for dom_id in summary.keys():
                if skipped_frames is not None and dom_id in skipped_frames[frame_index]:
                    continue
                rates[dom_id].append(summary[dom_id]["rates"])

        median_rates = {}
        for dom_id in rates.keys():
            median_rates[dom_id] = np.median(rates[dom_id], axis=0)

        return median_rates


class MedianPMTRatesService(kp.Module):
    def configure(self):
        self.rates = defaultdict(lambda: defaultdict(list))
        self.expose(self.get_median_rates, "GetMedianPMTRates")

    def process(self, blob):
        try:
            tmch_data = TMCHData(io.BytesIO(blob["CHData"]))
        except ValueError as e:
            self.log.error(e)
            self.log.error("Skipping corrupt monitoring channel packet.")
            return
        dom_id = tmch_data.dom_id
        for channel_id, rate in enumerate(tmch_data.pmt_rates):
            self.rates[dom_id][channel_id].append(rate)

    def get_median_rates(self):
        self.cprint("Calculating median PMT rates.")
        median_rates = {}
        for dom_id in self.rates.keys():
            median_rates[dom_id] = [np.median(self.rates[dom_id][c]) for c in range(31)]
        self.rates = defaultdict(lambda: defaultdict(list))
        return median_rates


class ResetTwofoldCounts(kp.Module):
    def process(self, blob):
        if "DumpTwofoldCounts" in self.services:
            self.cprint("Request twofold dump...")
            self.services["DumpTwofoldCounts"]()
        self.cprint("Resetting twofold counts")
        self.services["ResetTwofoldCounts"]()
        return blob


def calibrate_dom(
    dom_id,
    data,
    detector,
    livetime=None,
    fit_ang_dist=False,
    scale_mc_to_data=True,
    ad_fit_shape="pexp",
    fit_background=True,
    ctmin=-1.0,
):
    """Calibrate intra DOM PMT time offsets, efficiencies and sigmas

    Parameters
    ----------
    dom_id: DOM ID
    data: dict of coincidences or root or hdf5 file
    detector: instance of detector class
    livetime: data-taking duration [s]
    fixed_ang_dist: fixing angular distribution e.g. for data mc comparison
    auto_scale: auto scales the fixed angular distribution to the data

    Returns
    -------
    return_data: dictionary with fit results
    """

    if isinstance(data, str):
        filename = data
        loaders = {
            ".h5": load_k40_coincidences_from_hdf5,
            ".root": load_k40_coincidences_from_rootfile,
        }
        try:
            loader = loaders[os.path.splitext(filename)[1]]
        except KeyError:
            log.critical("File format not supported.")
            raise IOError
        else:
            data, livetime = loader(filename, dom_id)

    combs = np.array(list(combinations(range(31), 2)))
    angles = calculate_angles(detector, combs)
    cos_angles = np.cos(angles)
    angles = angles[cos_angles >= ctmin]
    data = data[cos_angles >= ctmin]
    combs = combs[cos_angles >= ctmin]

    try:
        fit_res = fit_delta_ts(data, livetime, fit_background=fit_background)
        rates, means, sigmas, popts, pcovs = fit_res
    except:
        return 0

    rate_errors = np.array([np.diag(pc)[2] for pc in pcovs])
    # mean_errors = np.array([np.diag(pc)[0] for pc in pcovs])
    scale_factor = None
    if fit_ang_dist:
        fit_res = fit_angular_distribution(
            angles, rates, rate_errors, shape=ad_fit_shape
        )
        fitted_rates, exp_popts, exp_pcov = fit_res
    else:
        mc_fitted_rates = exponential_polinomial(np.cos(angles), *MC_ANG_DIST)
        if scale_mc_to_data:
            scale_factor = np.mean(rates[angles < 1.5]) / np.mean(
                mc_fitted_rates[angles < 1.5]
            )
        else:
            scale_factor = 1.0
        fitted_rates = mc_fitted_rates * scale_factor
        exp_popts = []
        exp_pcov = []
        print("Using angular distribution from Monte Carlo")

    # t0_weights = np.array([0. if a>1. else 1. for a in angles])

    if not fit_background:
        minimize_weights = calculate_weights(fitted_rates, data)
    else:
        minimize_weights = fitted_rates

    opt_t0s = minimize_t0s(means, minimize_weights, combs)
    opt_sigmas = minimize_sigmas(sigmas, minimize_weights, combs)
    opt_qes = minimize_qes(fitted_rates, rates, minimize_weights, combs)
    corrected_means = correct_means(means, opt_t0s.x, combs)
    corrected_rates = correct_rates(rates, opt_qes.x, combs)
    rms_means, rms_corrected_means = calculate_rms_means(means, corrected_means)
    rms_rates, rms_corrected_rates = calculate_rms_rates(
        rates, fitted_rates, corrected_rates
    )
    cos_angles = np.cos(angles)
    return_data = {
        "opt_t0s": opt_t0s,
        "opt_qes": opt_qes,
        "data": data,
        "means": means,
        "rates": rates,
        "fitted_rates": fitted_rates,
        "angles": angles,
        "corrected_means": corrected_means,
        "corrected_rates": corrected_rates,
        "rms_means": rms_means,
        "rms_corrected_means": rms_corrected_means,
        "rms_rates": rms_rates,
        "rms_corrected_rates": rms_corrected_rates,
        "gaussian_popts": popts,
        "livetime": livetime,
        "exp_popts": exp_popts,
        "exp_pcov": exp_pcov,
        "scale_factor": scale_factor,
        "opt_sigmas": opt_sigmas,
        "sigmas": sigmas,
        "combs": combs,
    }
    return return_data


def calculate_weights(fitted_rates, data):
    comb_mean_rates = np.mean(data, axis=1)
    greater_zero = np.array(comb_mean_rates > 0, dtype=int)
    return fitted_rates * greater_zero


def load_k40_coincidences_from_hdf5(filename, dom_id):
    """Load k40 coincidences from hdf5 file

    Parameters
    ----------
    filename: filename of hdf5 file
    dom_id: DOM ID

    Returns
    -------
    data: numpy array of coincidences
    livetime: duration of data-taking
    """

    import h5py

    with h5py.File(filename, "r") as h5f:
        data = h5f["/k40counts/{0}".format(dom_id)]
        livetime = data.attrs["livetime"]
        data = np.array(data)

    return data, livetime


def load_k40_coincidences_from_rootfile(filename, dom_id):
    """Load k40 coincidences from JMonitorK40 ROOT file

    Parameters
    ----------
    filename: root file produced by JMonitorK40
    dom_id: DOM ID

    Returns
    -------
    data: numpy array of coincidences
    dom_weight: weight to apply to coincidences to get rate in Hz
    """

    from ROOT import TFile

    root_file_monitor = TFile(filename, "READ")
    dom_name = str(dom_id) + ".2S"
    histo_2d_monitor = root_file_monitor.Get(dom_name)
    data = []
    for c in range(1, histo_2d_monitor.GetNbinsX() + 1):
        combination = []
        for b in range(1, histo_2d_monitor.GetNbinsY() + 1):
            combination.append(histo_2d_monitor.GetBinContent(c, b))
        data.append(combination)

    weights = {}
    weights_histo = root_file_monitor.Get("weights_hist")
    try:
        for i in range(1, weights_histo.GetNbinsX() + 1):
            # we have to read all the entries, unfortunately
            weight = weights_histo.GetBinContent(i)
            label = weights_histo.GetXaxis().GetBinLabel(i)
            weights[label[3:]] = weight
        dom_weight = weights[str(dom_id)]
    except AttributeError:
        log.info("Weights histogram broken or not found, setting weight to 1.")
        dom_weight = 1.0
    return np.array(data), dom_weight


def gaussian(x, mean, sigma, rate, offset):
    return (
        rate / np.sqrt(2 * np.pi) / sigma * np.exp(-0.5 * (x - mean) ** 2 / sigma**2)
        + offset
    )


def gaussian_wo_offset(x, mean, sigma, rate):
    return (
        rate / np.sqrt(2 * np.pi) / sigma * np.exp(-0.5 * (x - mean) ** 2 / sigma**2)
    )


def fit_delta_ts(data, livetime, fit_background=True):
    """Fits gaussians to delta t for each PMT pair.

    Parameters
    ----------
    data: 2d np.array: x = PMT combinations (465), y = time, entry = frequency
    livetime: length of data taking in seconds
    fit_background: if True: fits gaussian with offset, else without offset

    Returns
    -------
    numpy arrays with rates and means for all PMT combinations
    """
    scipy = km3pipe.extras.scipy()
    from scipy import optimize

    data = data / livetime
    start = -(data.shape[1] - 1) / 2
    end = -start + 1
    xs = np.arange(start, end)

    rates = []
    sigmas = []
    means = []
    popts = []
    pcovs = []
    for combination in data:
        mean0 = np.argmax(combination) + start
        try:
            if fit_background:
                popt, pcov = optimize.curve_fit(
                    gaussian,
                    xs,
                    combination,
                    p0=[mean0, 4.0, 5.0, 0.1],
                    bounds=([start, 0, 0, 0], [end, 10, 10, 1]),
                )
            else:
                popt, pcov = optimize.curve_fit(
                    gaussian_wo_offset,
                    xs,
                    combination,
                    p0=[mean0, 4.0, 5.0],
                    bounds=([start, 0, 0], [end, 10, 10]),
                )
        except RuntimeError:
            popt = (0, 0, 0, 0)
        rates.append(popt[2])
        means.append(popt[0])
        sigmas.append(popt[1])
        popts.append(popt)
        pcovs.append(pcov)
    return (
        np.array(rates),
        np.array(means),
        np.array(sigmas),
        np.array(popts),
        np.array(pcovs),
    )


def calculate_angles(detector, combs):
    """Calculates angles between PMT combinations according to positions in
    detector_file

    Parameters
    ----------
    detector: detector description (kp.hardware.Detector)
    combs: pmt combinations

    Returns
    -------
    angles: numpy array of angles between all PMT combinations

    """
    angles = []
    pmt_angles = detector.pmt_angles
    for first, second in combs:
        angles.append(
            kp.math.angle_between(
                np.array(pmt_angles[first]), np.array(pmt_angles[second])
            )
        )
    return np.array(angles)


def exponential_polinomial(x, p1, p2, p3, p4):
    return 1 * np.exp(p1 + x * (p2 + x * (p3 + x * p4)))


def exponential(x, a, b):
    return a * np.exp(b * x)


def fit_angular_distribution(angles, rates, rate_errors, shape="pexp"):
    """Fits angular distribution of rates.

    Parameters
    ----------
    rates: numpy array
      with rates for all PMT combinations
    angles: numpy array
      with angles for all PMT combinations
    shape:
      which function to fit; exp for exponential or pexp for
      exponential_polinomial

    Returns
    -------
    fitted_rates: numpy array of fitted rates (fit_function(angles, popt...))

    """
    if shape == "exp":
        fit_function = exponential
        # p0 = [-0.91871169,  2.72224241, -1.19065965,  1.48054122]
    if shape == "pexp":
        fit_function = exponential_polinomial
        # p0 = [0.34921202, 2.8629577]

    cos_angles = np.cos(angles)
    popt, pcov = optimize.curve_fit(fit_function, cos_angles, rates)
    fitted_rates = fit_function(cos_angles, *popt)
    return fitted_rates, popt, pcov


def minimize_t0s(means, weights, combs):
    """Varies t0s to minimize the deviation of the gaussian means from zero.

    Parameters
    ----------
    means: numpy array of means of all PMT combinations
    weights: numpy array of weights for the squared sum
    combs: pmt combinations to use for minimization

    Returns
    -------
    opt_t0s: optimal t0 values for all PMTs

    """

    def make_quality_function(means, weights, combs):
        def quality_function(t0s):
            sq_sum = 0
            for mean, comb, weight in zip(means, combs, weights):
                sq_sum += ((mean - (t0s[comb[1]] - t0s[comb[0]])) * weight) ** 2
            return sq_sum

        return quality_function

    qfunc = make_quality_function(means, weights, combs)
    # t0s = np.zeros(31)
    t0s = np.random.rand(31)
    bounds = [(0, 0)] + [(-10.0, 10.0)] * 30
    opt_t0s = optimize.minimize(qfunc, t0s, bounds=bounds)
    return opt_t0s


def minimize_sigmas(sigmas, weights, combs):
    """Varies sigmas to minimize gaussian sigma12 - sqrt(sigma1² + sigma2²).

    Parameters
    ----------
    sigmas: numpy array of fitted sigmas of gaussians
    weights: numpy array of weights for the squared sum
    combs: pmt combinations to use for minimization

    Returns
    -------
    opt_sigmas: optimal sigma values for all PMTs

    """

    def make_quality_function(sigmas, weights, combs):
        def quality_function(s):
            sq_sum = 0
            for sigma, comb, weight in zip(sigmas, combs, weights):
                sigma_sqsum = np.sqrt(s[comb[1]] ** 2 + s[comb[0]] ** 2)
                sq_sum += ((sigma - sigma_sqsum) * weight) ** 2
            return sq_sum

        return quality_function

    qfunc = make_quality_function(sigmas, weights, combs)
    s = np.ones(31) * 2.5
    # s = np.random.rand(31)
    bounds = [(0.0, 5.0)] * 31
    opt_sigmas = optimize.minimize(qfunc, s, bounds=bounds)
    return opt_sigmas


def minimize_qes(fitted_rates, rates, weights, combs):
    """Varies QEs to minimize the deviation of the rates from the fitted_rates.

    Parameters
    ----------
    fitted_rates: numpy array of fitted rates from fit_angular_distribution
    rates: numpy array of rates of all PMT combinations
    weights: numpy array of weights for the squared sum
    combs: pmt combinations to use for minimization

    Returns
    -------
    opt_qes: optimal qe values for all PMTs

    """

    def make_quality_function(fitted_rates, rates, weights, combs):
        def quality_function(qes):
            sq_sum = 0
            for fitted_rate, comb, rate, weight in zip(
                fitted_rates, combs, rates, weights
            ):
                sq_sum += (
                    (rate / qes[comb[0]] / qes[comb[1]] - fitted_rate) * weight
                ) ** 2
            return sq_sum

        return quality_function

    qfunc = make_quality_function(fitted_rates, rates, weights, combs)
    qes = np.ones(31)
    bounds = [(0.1, 2.0)] * 31
    opt_qes = optimize.minimize(qfunc, qes, bounds=bounds)
    return opt_qes


def correct_means(means, opt_t0s, combs):
    """Applies optimal t0s to gaussians means.

    Should be around zero afterwards.

    Parameters
    ----------
    means: numpy array of means of gaussians of all PMT combinations
    opt_t0s: numpy array of optimal t0 values for all PMTs
    combs: pmt combinations used to correct

    Returns
    -------
    corrected_means: numpy array of corrected gaussian means for all PMT combs

    """
    corrected_means = np.array(
        [
            (opt_t0s[comb[1]] - opt_t0s[comb[0]]) - mean
            for mean, comb in zip(means, combs)
        ]
    )
    return corrected_means


def correct_rates(rates, opt_qes, combs):
    """Applies optimal qes to rates.

    Should be closer to fitted_rates afterwards.

    Parameters
    ----------
    rates: numpy array of rates of all PMT combinations
    opt_qes: numpy array of optimal qe values for all PMTs
    combs: pmt combinations used to correct

    Returns
    -------
    corrected_rates: numpy array of corrected rates for all PMT combinations
    """

    corrected_rates = np.array(
        [rate / opt_qes[comb[0]] / opt_qes[comb[1]] for rate, comb in zip(rates, combs)]
    )
    return corrected_rates


def calculate_rms_means(means, corrected_means):
    """Calculates RMS of means from zero before and after correction

    Parameters
    ----------
    means: numpy array of means of gaussians of all PMT combinations
    corrected_means: numpy array of corrected gaussian means for all PMT combs

    Returns
    -------
    rms_means: RMS of means from zero
    rms_corrected_means: RMS of corrected_means from zero
    """
    rms_means = np.sqrt(np.mean((means - 0) ** 2))
    rms_corrected_means = np.sqrt(np.mean((corrected_means - 0) ** 2))
    return rms_means, rms_corrected_means


def calculate_rms_rates(rates, fitted_rates, corrected_rates):
    """Calculates RMS of rates from fitted_rates before and after correction

    Parameters
    ----------
    rates: numpy array of rates of all PMT combinations
    corrected_rates: numpy array of corrected rates for all PMT combinations

    Returns
    -------
    rms_rates: RMS of rates from fitted_rates
    rms_corrected_rates: RMS of corrected_ratesrates from fitted_rates
    """
    rms_rates = np.sqrt(np.mean((rates - fitted_rates) ** 2))
    rms_corrected_rates = np.sqrt(np.mean((corrected_rates - fitted_rates) ** 2))
    return rms_rates, rms_corrected_rates


@jit(nopython=True)
def get_comb_index(i, j):
    """Return the index of PMT pair combinations"""
    return i * 30 - i * (i + 1) // 2 + j - 1


@jit(nopython=True)
def add_to_twofold_matrix(times, tdcs, mat, tmax=10):
    """Add counts to twofold coincidences for a given `tmax`.

    Parameters
    ----------
    times: np.ndarray of hit times (int32)
    tdcs: np.ndarray of channel_ids (uint8)
    mat: ref to a np.array((465, tmax * 2 + 1))
    tmax: int (time window)

    Returns
    -------
    mat: coincidence matrix (np.array((465, tmax * 2 + 1)))

    """
    h_idx = 0  # index of initial hit
    c_idx = 0  # index of coincident candidate hit
    n_hits = len(times)
    multiplicity = 0
    while h_idx <= n_hits:
        c_idx = h_idx + 1
        if (c_idx < n_hits) and (times[c_idx] - times[h_idx] <= tmax):
            multiplicity = 2
            c_idx += 1
            while (c_idx < n_hits) and (times[c_idx] - times[h_idx] <= tmax):
                c_idx += 1
                multiplicity += 1
            if multiplicity != 2:
                h_idx = c_idx
                continue
            c_idx -= 1
            h_tdc = tdcs[h_idx]
            c_tdc = tdcs[c_idx]
            h_time = times[h_idx]
            c_time = times[c_idx]
            if h_tdc != c_tdc:
                dt = int(c_time - h_time)
                if h_tdc > c_tdc:
                    mat[get_comb_index(c_tdc, h_tdc), -dt + tmax] += 1
                else:
                    mat[get_comb_index(h_tdc, c_tdc), dt + tmax] += 1
        h_idx = c_idx


# jmonitork40_comb_indices =  \
#     np.array((254, 423, 424, 391, 392, 255, 204, 205, 126, 120, 121, 0,
#     22, 12, 80, 81, 23, 48, 49, 148, 150, 96, 296, 221, 190, 191, 297, 312,
#     313, 386, 355, 132, 110, 431, 42, 433, 113, 256, 134, 358, 192, 74,
#     176, 36, 402, 301, 270, 69, 384, 2, 156, 38, 178, 70, 273, 404, 302,
#     77, 202, 351, 246, 440, 133, 262, 103, 118, 44, 141, 34, 4, 64, 30,
#     196, 91, 172, 61, 292, 84, 157, 198, 276, 182, 281, 410, 381, 289,
#     405, 439, 247, 356, 102, 263, 119, 140, 45, 35, 88, 65, 194, 31,
#     7, 60, 173, 82, 294, 158, 409, 277, 280, 183, 200, 288, 382, 406,
#     212, 432, 128, 388, 206, 264, 105, 72, 144, 52, 283, 6, 19, 14,
#     169, 24, 310, 97, 379, 186, 218, 59, 93, 152, 317, 304, 111, 387,
#     129, 207, 104, 265, 73, 18, 53, 5, 284, 146, 168, 15, 308, 26,
#     98, 92, 187, 58, 219, 380, 316, 154, 305, 112, 434, 257, 357, 135,
#     193, 300, 177, 401, 37, 75, 68, 271, 1, 385, 159, 403, 179, 272,
#     71, 39, 76, 303, 203, 213, 393, 248, 442, 298, 145, 184, 89, 377,
#     315, 216, 57, 309, 27, 99, 8, 54, 16, 171, 287, 153, 21, 78,
#     394, 441, 249, 299, 314, 185, 376, 90, 147, 56, 217, 25, 311, 100,
#     286, 55, 170, 17, 9, 20, 155, 79, 425, 426, 383, 306, 220, 290,
#     291, 307, 188, 189, 149, 151, 101, 86, 13, 50, 51, 87, 28, 29,
#     3, 352, 399, 375, 274, 407, 197, 285, 180, 279, 83, 295, 160, 199,
#     66, 174, 63, 33, 10, 95, 40, 400, 282, 275, 195, 408, 378, 278,
#     181, 293, 85, 161, 32, 67, 62, 175, 201, 94, 11, 41, 435, 415,
#     359, 360, 436, 347, 348, 258, 259, 318, 136, 162, 222, 223, 137, 114,
#     115, 43, 451, 443, 266, 389, 335, 456, 208, 396, 363, 250, 238, 327,
#     235, 107, 130, 215, 116, 343, 344, 452, 461, 462, 331, 332, 417, 226,
#     324, 371, 372, 229, 240, 241, 163, 142, 267, 230, 412, 122, 428, 319,
#     353, 227, 340, 166, 47, 108, 253, 138, 444, 411, 231, 427, 123, 320,
#     46, 228, 165, 341, 354, 252, 109, 139, 455, 336, 395, 209, 364, 106,
#     239, 234, 328, 251, 214, 131, 117, 373, 447, 243, 418, 164, 369, 325,
#     460, 342, 329, 237, 224, 242, 448, 419, 339, 370, 459, 326, 167, 236,
#     330, 225, 127, 365, 124, 333, 244, 450, 430, 397, 211, 260, 366, 429,
#     334, 449, 245, 125, 210, 398, 261, 321, 420, 421, 422, 322, 367, 368,
#     323, 345, 413, 232, 143, 268, 446, 361, 463, 464, 346, 453, 454, 416,
#     374, 233, 337, 458, 349, 414, 457, 338, 350, 445, 269, 362, 390, 437,
#     438))
#
"""
jmonitork40_comb_indices =  \
    np.array((417, 418, 419, 420, 421, 422, 363, 364, 365, 366, 367, 368,
    318, 319, 320, 321, 322, 323, 156, 157, 158, 159, 160, 161, 96, 97, 98,
    99, 100, 101, 461, 369, 324, 371, 464, 427, 331, 237, 238, 333, 434, 415,
    339, 231, 175, 232, 342, 278, 184, 61, 62, 186, 281, 220, 162, 54, 13,
    56, 168, 459, 370, 325, 374, 423, 428, 328, 244, 239, 338, 343, 411, 346,
    226, 178, 229, 270, 271, 181, 68, 69, 191, 170, 216, 164, 50, 16, 58,
    462, 373, 326, 327, 429, 424, 337, 240, 245, 222, 345, 412, 347, 228, 179,
    180, 272, 273, 190, 70, 71, 48, 163, 217, 172, 57, 17, 463, 372, 234,
    332, 430, 431, 334, 241, 174, 230, 340, 416, 341, 233, 60, 185, 279, 280,
    187, 63, 12, 52, 165, 221, 166, 59, 460, 242, 235, 336, 425, 432, 330,
    223, 176, 225, 348, 413, 350, 64, 65, 189, 274, 275, 183, 49, 14, 55,
    173, 218, 169, 335, 236, 243, 329, 433, 426, 344, 224, 177, 227, 349, 414,
    188, 66, 67, 182, 276, 277, 171, 53, 15, 51, 167, 219, 387, 204, 128,
    209, 396, 443, 435, 263, 112, 120, 249, 375, 283, 73, 6, 85, 301, 310,
    306, 148, 22, 28, 154, 388, 208, 126, 211, 254, 451, 452, 256, 104, 105,
    282, 383, 285, 84, 1, 87, 144, 312, 313, 150, 24, 25, 395, 210, 129,
    110, 262, 436, 445, 248, 121, 72, 284, 376, 300, 86, 7, 18, 146, 307,
    314, 152, 29, 389, 205, 111, 118, 247, 446, 437, 265, 4, 81, 299, 377,
    287, 75, 19, 26, 149, 315, 308, 155, 390, 255, 102, 103, 257, 453, 454,
    80, 0, 83, 286, 384, 289, 145, 20, 21, 151, 316, 317, 444, 246, 119,
    113, 264, 438, 298, 82, 5, 74, 288, 378, 311, 147, 27, 23, 153, 309,
    351, 136, 42, 138, 354, 403, 194, 33, 34, 200, 410, 385, 292, 91, 3,
    94, 297, 359, 137, 44, 133, 399, 404, 196, 40, 35, 202, 290, 379, 303,
    92, 10, 79, 352, 132, 45, 192, 405, 400, 198, 36, 41, 88, 302, 380,
    294, 78, 11, 353, 139, 30, 195, 406, 407, 201, 37, 2, 90, 293, 386,
    296, 95, 360, 38, 31, 197, 401, 408, 203, 89, 8, 77, 295, 381, 305,
    193, 32, 39, 199, 409, 402, 291, 76, 9, 93, 304, 382, 355, 134, 46,
    141, 362, 455, 439, 251, 108, 124, 269, 356, 140, 43, 143, 258, 447, 448,
    260, 116, 117, 361, 142, 47, 106, 250, 440, 457, 268, 125, 357, 135, 107,
    122, 267, 458, 441, 253, 358, 259, 114, 115, 261, 449, 450, 456, 266, 123,
    109, 252, 442, 391, 212, 127, 214, 394, 397, 213, 130, 207, 392, 206, 131,
    393, 215, 398))
"""



# Filename: ahrs.py
# -*- coding: utf-8 -*-
# pylint: disable=locally-disabled
"""
AHRS calibration.

"""

import io
from collections import defaultdict
import time
import xml.etree.ElementTree as ET

import km3db

import numpy as np
from numpy import cos, sin, arctan2
import km3pipe as kp
from km3pipe.tools import timed_cache
from km3pipe.io.daq import TMCHData

__author__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

log = kp.logger.get_logger(__name__)  # pylint: disable=C0103

# log.setLevel("DEBUG")


class AHRSCalibrator(kp.Module):
    """Calculates AHRS yaw, pitch and roll from median A and H of an interval.

    Parameters
    ----------
    det_id: int
        The detector ID, e.g. 29)

    Other Parameters
    ----------------
    interval: int (accumulation interval in [sec], default: 10s)

    Notes
    -----
    Writes 'AHRSCalibration' in the blob with:
        dict: key=dom_id, value=tuple: (timestamp, du, floor, yaw, pitch, roll)

    """

    def configure(self):
        det_id = self.require("det_id")
        self.interval = self.get("interval") or 10  # in sec
        self.A = defaultdict(list)
        self.H = defaultdict(list)
        self.detector = kp.hardware.Detector(det_id=det_id)
        self.clbmap = km3db.CLBMap(det_id)
        self.timestamp = time.time()

    def process(self, blob):
        tmch_data = TMCHData(io.BytesIO(blob["CHData"]))
        dom_id = tmch_data.dom_id
        try:
            du, floor, _ = self.detector.doms[dom_id]
        except KeyError:  # base CLB
            return blob

        self.A[dom_id].append(tmch_data.A)
        self.H[dom_id].append(tmch_data.H)

        if time.time() - self.timestamp > self.interval:
            self.timestamp = time.time()
            calib = self.calibrate()
            blob["AHRSCalibration"] = calib

        return blob

    def calibrate(self):
        """Calculate yaw, pitch and roll from the median of A and H.

        After successful calibration, the `self.A` and `self.H` are reset.
        DOMs with missing AHRS pre-calibration data are skipped.

        Returns
        -------
        dict: key=dom_id, value=tuple: (timestamp, du, floor, yaw, pitch, roll)

        """
        now = time.time()
        dom_ids = self.A.keys()
        print("Calibrating AHRS from median A and H for {} DOMs.".format(len(dom_ids)))
        calibrations = {}
        for dom_id in dom_ids:
            print("Calibrating DOM ID {}".format(dom_id))
            clb_upi = self.clbmap.doms_ids[dom_id].clb_upi
            ahrs_calib = get_latest_ahrs_calibration(clb_upi)
            if ahrs_calib is None:
                log.warning("AHRS calibration missing for '{}'".format(dom_id))
                continue
            du, floor, _ = self.detector.doms[dom_id]
            A = np.median(self.A[dom_id], axis=0)
            H = np.median(self.H[dom_id], axis=0)

            cyaw, cpitch, croll = fit_ahrs(A, H, *ahrs_calib)
            calibrations[dom_id] = (now, du, floor, cyaw, cpitch, croll)
        self.A = defaultdict(list)
        self.H = defaultdict(list)
        return calibrations


def fit_ahrs(A, H, Aoff, Arot, Hoff, Hrot):
    """Calculate yaw, pitch and roll for given A/H and calibration set.

    Author: Vladimir Kulikovskiy

    Parameters
    ----------
    A: list, tuple or numpy.array of shape (3,)
    H: list, tuple or numpy.array of shape (3,)
    Aoff: numpy.array of shape(3,)
    Arot: numpy.array of shape(3, 3)
    Hoff: numpy.array of shape(3,)
    Hrot: numpy.array of shape(3, 3)

    Returns
    -------
    yaw, pitch, roll

    """
    Acal = np.dot(A - Aoff, Arot)
    Hcal = np.dot(H - Hoff, Hrot)

    # invert axis for DOM upside down
    for i in (1, 2):
        Acal[i] = -Acal[i]
        Hcal[i] = -Hcal[i]

    roll = arctan2(-Acal[1], -Acal[2])
    pitch = arctan2(Acal[0], np.sqrt(Acal[1] * Acal[1] + Acal[2] * Acal[2]))
    yaw = arctan2(
        Hcal[2] * sin(roll) - Hcal[1] * cos(roll),
        sum(
            (
                Hcal[0] * cos(pitch),
                Hcal[1] * sin(pitch) * sin(roll),
                Hcal[2] * sin(pitch) * cos(roll),
            )
        ),
    )

    yaw = np.degrees(yaw)
    while yaw < 0:
        yaw += 360
    # yaw = (yaw + magnetic_declination + 360 ) % 360
    roll = np.degrees(roll)
    pitch = np.degrees(pitch)
    return yaw, pitch, roll


@timed_cache(hours=1, maxsize=None, typed=False)
def get_latest_ahrs_calibration(clb_upi, max_version=3):
    """Retrieve the latest AHRS calibration data for a given CLB

    Parameters
    ----------
    clb_upi: str
    max_version: int, maximum version to check, optional

    Returns
    -------
    Aoff: numpy.array with shape(3,)
    Arot: numpy.array with shape(3,3)
    Hoff: numpy.array with shape(3,)
    Hrot: numpy.array with shape(3,3)

    or None if no calibration found.

    """
    ahrs_upi = km3db.tools.clbupi2compassupi(clb_upi)

    db = km3db.DBManager()

    datasets = []
    for version in range(max_version, 0, -1):
        for n in range(1, 100):
            log.debug("Iteration #{} to get the calib data".format(n))
            url = (
                "show_product_test.htm?upi={0}&"
                "testtype=AHRS-CALIBRATION-v{1}&n={2}&out=xml".format(
                    ahrs_upi, version, n
                )
            )
            log.debug("AHRS calib DB URL: {}".format(url))
            _raw_data = db.get(url).replace("\n", "")
            log.debug("What I got back as AHRS calib: {}".format(_raw_data))
            if len(_raw_data) == 0:
                break
            try:
                xroot = ET.parse(io.StringIO(_raw_data)).getroot()
            except ET.ParseError:
                continue
            else:
                datasets.append(xroot)

    if len(datasets) == 0:
        return None

    latest_dataset = _get_latest_dataset(datasets)
    return _extract_calibration(latest_dataset)


def _get_latest_dataset(datasets):
    """Find the latest valid AHRS calibration dataset"""
    return sorted(datasets, key=lambda d: d.findall(".//EndTime")[0].text)[-1]


def _extract_calibration(xroot):
    """Extract AHRS calibration information from XML root.

    Parameters
    ----------
    xroot: XML root


    Returns
    -------
    Aoff: numpy.array with shape(3,)
    Arot: numpy.array with shape(3,3)
    Hoff: numpy.array with shape(3,)
    Hrot: numpy.array with shape(3,3)

    """
    names = [c.text for c in xroot.findall(".//Name")]
    val = [[i.text for i in c] for c in xroot.findall(".//Values")]

    # The fields has to be reindeced, these are the index mappings
    col_ic = [int(v) for v in val[names.index("AHRS_Matrix_Column")]]
    try:
        row_ic = [int(v) for v in val[names.index("AHRS_Matrix_Row")]]
    except ValueError:
        row_ic = [2, 2, 2, 1, 1, 1, 0, 0, 0]
    try:
        vec_ic = [int(v) for v in val[names.index("AHRS_Vector_Index")]]
    except ValueError:
        vec_ic = [2, 1, 0]

    Aoff_ix = names.index("AHRS_Acceleration_Offset")
    Arot_ix = names.index("AHRS_Acceleration_Rotation")
    Hrot_ix = names.index("AHRS_Magnetic_Rotation")

    Aoff = np.array(val[Aoff_ix])[vec_ic].astype(float)
    Arot = (
        np.array(val[Arot_ix]).reshape(3, 3)[col_ic, row_ic].reshape(3, 3).astype(float)
    )
    Hrot = (
        np.array(val[Hrot_ix]).reshape(3, 3)[col_ic, row_ic].reshape(3, 3).astype(float)
    )

    Hoff = []
    for q in "XYZ":
        values = []
        for t in ("Min", "Max"):
            ix = names.index("AHRS_Magnetic_{}{}".format(q, t))
            values.append(float(val[ix][0]))
        Hoff.append(sum(values) / 2.0)
    Hoff = np.array(Hoff)

    return Aoff, Arot, Hoff, Hrot



#!/usr/bin/env python
# Filename: triggermap.py
# vim: ts=4 sw=4 et
"""
This script creates histogram which shows the trigger contribution for events.

Usage:
    triggermap [options] -d DET_ID_OR_DETX FILENAMES...
    triggermap --version

Option:
    FILENAMES          Name of the input file(s).
    --offline          Read offline events instead.
    -u DU              Only plot for the given DU.
    -d DET_ID_OR_DETX  Detector ID or DETX file.
    -p PLOT_FILENAME   The filename of the plot [default: trigger_map.png].
    -s SUBTITLE        Optional subtitle for the plot instead of the filename.
    -h --help          Show this screen.

"""

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

from docopt import docopt
import matplotlib

# Force matplotlib to not use any Xwindows backend.
matplotlib.use("Agg")  # noqa
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
import numpy as np

import km3pipe as kp
from km3modules.common import StatusBar
import km3modules as km
import km3pipe.style

km3pipe.style.use("km3pipe")


class TriggerMap(kp.Module):
    """Creates a plot to show the number of triggered hits for each DOM."""

    def configure(self):
        self.det = self.require("detector")
        self.plot_filename = self.require("plot_filename")
        self.subtitle = self.get("subtitle", default="")
        self.du = self.get("du")
        if self.du is not None:
            self.n_dus = 1
            self.n_doms = 18
            self.dus = [self.du]
        else:
            self.n_dus = self.det.n_dus
            self.n_doms = int(self.det.n_doms / self.n_dus)
            self.dus = sorted(self.det.dus)
        self.n_rows = self.n_dus * self.n_doms
        self.hit_counts = []

    def process(self, blob):
        hits = blob["Hits"]
        hits = hits[hits.triggered.astype(bool)]
        dom_ids = np.unique(hits.dom_id)
        hit_counts = np.zeros(self.n_dus * self.n_doms)
        for dom_id in dom_ids:
            n_hits = np.sum(hits.dom_id == dom_id)
            du, floor, _ = self.det.doms[dom_id]
            if self.du is not None and du != self.du:
                continue
            hit_counts[(self.dus.index(du) - 1) * self.n_doms + floor - 1] += n_hits
        self.hit_counts.append(hit_counts)

        return blob

    def finish(self):
        self.create_plot()

    def create_plot(self):
        title_du = " - DU {}".format(self.du) if self.du else ""
        title = "Trigger Map{}\n{}".format(title_du, self.subtitle)
        fig, ax = plt.subplots(figsize=(16, 8))
        ax.grid(True)
        ax.set_axisbelow(True)
        hit_mat = np.array([np.array(x) for x in self.hit_counts]).transpose()
        im = ax.matshow(
            hit_mat,
            interpolation="nearest",
            filternorm=None,
            cmap="plasma",
            aspect="auto",
            origin="lower",
            zorder=3,
            norm=LogNorm(vmin=1, vmax=np.amax(hit_mat)),
        )
        yticks = np.arange(self.n_rows)
        floors_to_label = range(self.n_doms) if self.n_dus == 1 else [1, 6, 12]
        ytick_labels = [
            "DU{}-DOM{}".format(du, floor) if floor in floors_to_label else ""
            for (du, floor, _) in self.det.doms.values()
        ]
        ax.set_yticks(yticks)
        ax.set_yticklabels(ytick_labels)
        ax.tick_params(labelbottom=True)
        ax.tick_params(labeltop=True)
        ax.set_xlabel("event id")
        ax.set_title(title)
        cb = fig.colorbar(im, pad=0.05)
        cb.set_label("number of triggered hits")

        fig.tight_layout()

        plt.savefig(self.plot_filename, dpi=120, bbox_inches="tight")


def main():
    args = docopt(__doc__, version=kp.version)

    du = int(args["-u"]) if args["-u"] else None

    try:
        det_id = int(args["-d"])
        det = kp.hardware.Detector(det_id=det_id)
    except ValueError:
        detx = args["-d"]
        det = kp.hardware.Detector(filename=detx)

    if args["-s"] is not None:
        subtitle = args["-s"]
    else:
        subtitle = ", ".join(args["FILENAMES"])

    pipe = kp.Pipeline()
    pipe.attach(
        km.common.MultiFilePump, pump=kp.io.OfflinePump, filenames=args["FILENAMES"]
    )
    pipe.attach(km.io.HitsTabulator, kind="offline")
    pipe.attach(StatusBar, every=2500)
    pipe.attach(
        TriggerMap,
        detector=det,
        du=du,
        plot_filename=args["-p"],
        subtitle=subtitle,
    )
    pipe.drain()


if __name__ == "__main__":
    main()



# Filename: app.py
"""
PipeInspector

Usage:
    pipeinspector FILE
    pipeinspector (-h | --help)
    pipeinspector --version

Options:
    -h --help       Show this screen.

"""

import os

from pipeinspector.settings import UI
import km3pipe.extras
from km3pipe.io import OfflinePump, HDF5Pump, EvtPump, CLBPump
from km3pipe.io.daq import DAQPump

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"
__version__ = "1.0.0"


def handle_input(input):
    """Handle any unhandled input."""
    if input in UI.keys["escape"]:
        raise urwid.ExitMainLoop


def filter_input(keys, raw):
    """Adds fancy mouse wheel functionality and VI navigation to ListBox"""
    if len(keys) == 1:
        if keys[0] in UI.keys["up"]:
            keys[0] = "up"
        elif keys[0] in UI.keys["down"]:
            keys[0] = "down"
        elif len(keys[0]) == 4 and keys[0][0] == "mouse press":
            if keys[0][1] == 4:
                keys[0] = "up"
            elif keys[0][1] == 5:
                keys[0] = "down"
    return keys


def get_pump(input_file):
    extension = os.path.splitext(input_file)[1][1:]
    if extension == "evt":
        pump = EvtPump(filename=input_file, cache_enabled=True)
    elif extension == "dat":
        pump = DAQPump(filename=input_file)
    elif extension == "dqd":
        pump = CLBPump(filename=input_file, cache_enabled=True)
    elif extension == "root":
        pump = OfflinePump(filename=input_file)
    elif extension == "h5":
        pump = HDF5Pump(filename=input_file)
    else:
        raise SystemExit("No pump found for '{0}' files.".format(extension))
    return pump


def main():
    from docopt import docopt

    arguments = docopt(__doc__, version=__version__)

    urwid = km3pipe.extras.urwid()

    input_file = arguments["FILE"]
    pump = get_pump(input_file)

    from pipeinspector.gui import MainFrame

    main_frame = MainFrame(pump)
    # main_frame.header.set_text("Inspecting {0}".format(input_file))

    loop = urwid.MainLoop(
        main_frame, UI.palette, input_filter=filter_input, unhandled_input=handle_input
    )
    loop.run()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# coding: utf-8 -*-
"""

==============
Sphere Cuts
==============

This example shows some spher(ical) cuts on a set of points in 3D
using ``kp.math.spherecut`` and ``kp.math.spherecutmask``.

"""
import km3pipe as kp
import matplotlib.pyplot as plt
import numpy as np


#######################
# Generating a few points randomly distributed in a cube

points = np.random.rand(1000, 3)

#######################
# In this example we create two masks and use them to not only highlight the
# selected points but also to hide those from the original dataset.

mask1 = kp.math.spherecutmask([0, 0, 0], 0.8, 1.0, points)
mask2 = kp.math.spherecutmask([0.8, 0.8, 0.8], 0, 0.2, points)

fig = plt.figure()
ax = plt.axes(projection="3d")
ax.scatter(*points[~(mask1 | mask2)].T, label="unselected points")
ax.scatter(*points[mask1].T, label="mask1")
ax.scatter(*points[mask2].T, label="mask2")

ax.legend()

#######################
# This will select all points which are inside a sphere with radius 0.5
# and centered at (0.5, 0.5, 0.5):

selected_points = kp.math.spherecut([0.5, 0.5, 0.5], 0.0, 0.5, points)

fig = plt.figure()
ax = plt.axes(projection="3d")

ax.scatter(*selected_points.T)

fig.tight_layout()
plt.show()
